"""err.py: All of RAIMAD errors in one conveniet module."""

from raimad.compo import InvalidSubcompoError
from raimad.compo import CompoInsteadOfProxyAsSubcompoError
from raimad.compo import TransformCompoError
from raimad.compo import CopyCompoError
from raimad.compo import ProxyCompoConfusionError

from raimad.ansec import AnSecError
from raimad.ansec import AnSecRadiusError
from raimad.ansec import AnSecRadiusTooManyArgumentsError
from raimad.ansec import AnSecRadiusNotEnoughArgumentsError
from raimad.ansec import AnSecRadiusIncorrectArgumentsError
from raimad.ansec import AnSecThetaError
from raimad.ansec import AnSecThetaTooManyArgumentsError
from raimad.ansec import AnSecThetaNotEnoughArgumentsError
from raimad.ansec import AnSecThetaIncorrectArgumentsError

from raimad.bbox import EmptyBBoxError

from raimad.cif.shorthand import InvalidDestinationError

from raimad.string_import import StringImportError

__all__ = [
    "InvalidSubcompoError",
    "CompoInsteadOfProxyAsSubcompoError",
    "TransformCompoError",
    "CopyCompoError",
    "ProxyCompoConfusionError",
    "AnSecError",
    "AnSecRadiusError",
    "AnSecRadiusTooManyArgumentsError",
    "AnSecRadiusNotEnoughArgumentsError",
    "AnSecRadiusIncorrectArgumentsError",
    "AnSecThetaError",
    "AnSecThetaTooManyArgumentsError",
    "AnSecThetaNotEnoughArgumentsError",
    "AnSecThetaIncorrectArgumentsError",
    "EmptyBBoxError",
    "InvalidDestinationError",
    "StringImportError",
    ]




import unittest
from math import radians, sqrt

import raimad as rai
from raimad import add, eq, sub, midpoint

from .utils import ArrayAlmostEqual


class TestPolys(ArrayAlmostEqual, unittest.TestCase):

    def test_angle_between(self):
        self.assertAlmostEqual(
            rai.angle_between((0, 0), (10, 0)),
            radians(0))

        self.assertAlmostEqual(
            rai.angle_between((0, 0), (100, 0)),
            radians(0))

        self.assertAlmostEqual(
            rai.angle_between((10, 0), (100, 0)),
            radians(0))

        self.assertAlmostEqual(
            rai.angle_between((10, 0), (-100, 0)),
            radians(180))

        self.assertAlmostEqual(
            rai.angle_between((10, 10), (20, 20)),
            radians(45))

    def test_polar(self):
        self.assertArrayAlmostEqual(
            rai.polar(radians(60), 2),
            (1, sqrt(3))
            )

        self.assertArrayAlmostEqual(
            rai.polar(radians(-60), 2),
            (1, -sqrt(3))
            )

        self.assertArrayAlmostEqual(
            rai.polar(radians(180), 100),
            (-100, 0)
            )

    def test_is_rotated(self):
        self.assertTrue(rai.iters.is_rotated(
            'abc',
            'bca',
            ))

        self.assertFalse(rai.iters.is_rotated(
            'abc',
            'aabc',
            ))

        self.assertFalse(rai.iters.is_rotated(
            'abc',
            'cba',
            ))

        self.assertTrue(rai.iters.is_rotated(
            'abcd',
            'bcda',
            ))

        self.assertTrue(rai.iters.is_rotated(
            [1, 2, 3, 4],
            [3, 4, 1, 2],
            ))

        self.assertTrue(rai.iters.is_rotated(
            [
                (1, 2),
                (3, 4),
                (5, 6),
                (7, 8),
                ],
            [
                (5, 6),
                (7, 8),
                (1, 2),
                (3, 4),
                ],
            ))

        self.assertFalse(rai.iters.is_rotated(
            [
                (1, 2),
                (3, 4),
                (5, 6),
                (7, 8),
                ],
            [
                (1, 2),
                (3, 4),
                (7, 8),
                (5, 6),
                ],
            ))

    def test_flatten(self):
        self.assertEqual(
            rai.flatten([[[1, 2], 3, [4], [5, 6], 7], 8]),
            [1, 2, 3, 4, 5, 6, 7, 8]
            )

        self.assertEqual(
            rai.flatten([1, 2, 3, 4, 5, 6, 7, 8]),
            [1, 2, 3, 4, 5, 6, 7, 8]
            )

        self.assertEqual(
            rai.flatten([[['a', 'bcd'], 'e', ['f'], ['g', 'h'], 'ij'], 'klm']),
            ['a', 'bcd', 'e', 'f', 'g', 'h', 'ij', 'klm']
            )

        self.assertEqual(
            rai.flatten('string'),
            ['string'],
            )

        self.assertEqual(
            rai.flatten(10),
            [10],
            )

    def test_add_eq(self):
        self.assertTrue(
            rai.eq(
                rai.add(
                    (10, 12),
                    (13, 14),
                    ),
                (23, 26)
                )
            )

    def test_add_eq_infix(self):
        self.assertTrue(
            (23, 26) |eq| ( (10, 12) |add| (13, 14) )
            )

    def test_sub_eq(self):
        self.assertTrue(
            rai.eq(
                rai.sub(
                    (10, 12),
                    (13, 14),
                    ),
                (-3, -2)
                )
            )

    def test_sub_infix(self):
        self.assertTrue(
            (-3, -2) |eq| ( (10, 12) |sub| (13, 14) )
            )

    def test_midpoint(self):
        self.assertTrue(
            rai.eq(
                rai.midpoint((10, 20), (20, 40)),
                (15, 30)
                )
            )

    def test_midpoint_infix(self):
        self.assertTrue(
            rai.eq(
                (10, 20) |midpoint| (20, 40),
                (15, 30)
                )
            )

    def distance_between(self):
        self.assertEqual(
            rai.distance_between((10, 9), (7, 5)),
            5
            )


    #def test_force_evaluate(self):
    #    # We define it here as a factory, because
    #    # the various checks, etc.
    #    # alter the state of a generator
    #    def generator_of_maps():
    #        return (
    #            map(lambda x: x ** 2, range(x))
    #            for x in range(5)
    #            )

    #    self.assertTrue(isinstance(generator_of_maps(), types.GeneratorType))
    #    self.assertTrue(isinstance(
    #        next(generator_of_maps()),
    #        map
    #        ))

    #    evaluated_as_lists = rai.force_evaluate(generator_of_maps(), list)
    #    evaluated_as_tuples = rai.force_evaluate(generator_of_maps(), tuple)

    #    self.assertEqual(
    #        evaluated_as_lists,
    #        [
    #            [0, ],
    #            [0, 1, ],
    #            [0, 1, 2, ],
    #            [0, 1, 4, 9,],
    #            [0, 1, 4, 9, 16, ],
    #            ]
    #        )

    #    self.assertEqual(
    #        evaluated_as_tuples,
    #        (
    #            (0, ),
    #            (0, 1, ),
    #            (0, 1, 2, ),
    #            (0, 1, 4, 9,),
    #            (0, 1, 4, 9, 16, ),
    #            )
    #        )


if __name__ == '__main__':
    unittest.main()




"""helpers.py: misc helper functions."""

from typing import Iterator, TypeVar, Callable, Generic
import math

try:
    from typing import Self
except ImportError:
    # py3.10 and lower
    from typing_extensions import Self

import raimad as rai

fullcircle = math.radians(360)
halfcircle = math.radians(180)
quartercircle = math.radians(90)
eigthcircle = math.radians(45)

semicircle = math.radians(180)
demisemicircle = math.radians(90)
hemidemisemicircle = math.radians(45)

def angle_between(p1: 'rai.typing.Point', p2: 'rai.typing.Point') -> float:
    """Angle between two points."""
    x = p2[0] - p1[0]
    y = p2[1] - p1[1]
    return math.atan2(y, x)

def polar(arg: float, mod: float = 1) -> 'rai.typing.Point':
    """Construct an XY point from an argument and modulus."""
    return (
        math.cos(arg) * mod,
        math.sin(arg) * mod
        )

def is_compo_class(obj: type) -> bool:
    """
    Check whether an object is a CompoType.

    Note that this returns False on the `rai.Compo`
    abstract base class.
    """
    return (
        isinstance(obj, type)
        and issubclass(obj, rai.Compo)
        and obj is not rai.Compo
        )

def _custom_base(value: int, glyphs: list[str]) -> Iterator[str]:
    """
    Encode an integer into base-N using a custom set of glyphs (generator).

    Parameters
    ----------
    value: int
        The integer to encode
    glyphs: list[str]
        A list of N strings representing the glyphs
        in your custom base.
        TODO low to high or high to low? I forgot.

    Yields
    ------
    str
        Your encoded integer is yielded, glyph by glyph
    """
    div, mod = divmod(value, len(glyphs))
    if div > 0:
        yield from _custom_base(div, glyphs)
    yield glyphs[mod]

def custom_base(value: int, glyphs: list[str]) -> str:
    """
    Encode an integer into base-N using a custom set of glyphs.

    Parameters
    ----------
    value: int
        The integer to encode
    glyphs: list[str]
        A list of N strings representing the glyphs
        in your custom base.
        TODO low to high or high to low? I forgot.

    Returns
    -------
    str
        A string with your encoded integer.
    """
    return ''.join(_custom_base(value, glyphs))


WINGDINGS = [
    f"\033[0;{color}m{symbol} "
    for color in [31, 32, 34, 35, 36]
    for symbol in r"●■◀▶▲▼▚x╚╔╝═╩╗╦╵╷┴┬╰/\╭╮╯∞o8:"
    ]

def wingdingify(value: int) -> str:
    """Encode an integer with a bunch of symbols."""
    return ''.join((
        custom_base(value, WINGDINGS),
        '\033[0m',
        ))

T = TypeVar('T')
R = TypeVar('R')
class Infix(Generic[T, R]):
    left: T | None

    def __init__(self) -> None:
        self.left = None

    def __ror__(self, left: T) -> Self:
        self.left = left
        return self

    def __or__(self, right: T) -> R:
        assert \
            self.left is not None, \
            "Incorrect usage of infix operator"

        return self(self.left, right)

    def __call__(self, left: T, right: T) -> R:
        raise NotImplemented

class Midpoint(Infix['rai.typing.PointLike', 'rai.typing.PointLike']):
    def __call__(
            self,
            p1: 'rai.typing.PointLike',
            p2: 'rai.typing.PointLike'
            ) -> 'rai.typing.PointLike':
        """Midpoint between two points."""
        return (
            (p1[0] + p2[0]) / 2,
            (p1[1] + p2[1]) / 2,
            )

midpoint = Midpoint()

class Add(Infix['rai.typing.PointLike', 'rai.typing.PointLike']):
    def __call__(
            self,
            p1: 'rai.typing.PointLike',
            p2: 'rai.typing.PointLike'
            ) -> 'rai.typing.PointLike':
        return (
            p1[0] + p2[0],
            p1[1] + p2[1],
            )

add = Add()

class Sub(Infix['rai.typing.PointLike', 'rai.typing.PointLike']):
    def __call__(
            self,
            p1: 'rai.typing.PointLike',
            p2: 'rai.typing.PointLike'
            ) -> 'rai.typing.PointLike':
        return (
            p1[0] - p2[0],
            p1[1] - p2[1],
            )

sub = Sub()

class Eq(Infix['rai.typing.PointLike', bool]):
    def __call__(
            self,
            p1: 'rai.typing.PointLike',
            p2: 'rai.typing.PointLike'
            ) -> bool:
        return rai.affine.norm((
            p1[0] - p2[0],
            p1[1] - p2[1],
            )) < rai.epsilon

eq = Eq()

def distance_between(
        p1: 'rai.typing.Point',
        p2: 'rai.typing.Point'
        ) -> float:

    return rai.affine.norm((
        p1[0] - p2[0],
        p1[1] - p2[1],
        ))




import unittest

import raimad as rai

class Annotated(rai.Compo):
    class Marks:
        square_corner = rai.Mark('corner of the square')
        square_center = rai.Mark('center of the square')

    class Layers:
        al = rai.Layer('Aluminium')
        insl = rai.Layer('Insulator')
        gnd = rai.Layer('Ground')

    class Options:
        width = rai.Option.Geometric('width of coupler')
        freq = rai.Option.Functional('resonant frequency')
        grav = rai.Option.Environmental('Gravitational constant')
        skip_beams = rai.Option.Debug('Do not generate beams')
        print_beams = rai.Option('Print a line for every beam')

    def _make(
            width,
            freq=3e9,
            grav: float = 6.67430e-11,
            skip_beams: bool = True,
            print_beams=False,
            ):
        pass


class TestClassIntrospection(unittest.TestCase):

    def test_class_introspection_marks(self):

        self.assertEqual(
            Annotated.Marks[0].name,
            'square_corner'
            )

        self.assertEqual(
            Annotated.Marks[1].desc,
            'center of the square'
            )

        self.assertEqual(
            Annotated.Marks['square_corner'].name,
            'square_corner'
            )

        self.assertEqual(
            Annotated.Marks['square_center'].desc,
            'center of the square'
            )

    def test_class_introspection_layers(self):

        self.assertEqual(
            Annotated.Layers[2].name,
            'gnd'
            )

        self.assertEqual(
            Annotated.Layers[2].desc,
            'Ground'
            )

    def test_class_introspection_options(self):

        self.assertEqual(
            Annotated.Options[3].name,
            'skip_beams'
            )

        self.assertEqual(
            Annotated.Options[4].desc,
            'Print a line for every beam'
            )

        self.assertTrue(
            Annotated.Options[3].category
            is
            rai.Option.Debug
            )

        self.assertTrue(
            Annotated.Options[4].category
            is
            rai.Option
            )

    def test_class_introspection_options_shorthands(self):

        self.assertTrue(Annotated.Options[3].default is True)
        self.assertTrue(Annotated.Options[4].default is False)

        # Type specified explicitly with an annotation
        self.assertEqual(Annotated.Options[3].annot, bool)
        # Type guessed from type of default
        self.assertEqual(Annotated.Options[4].annot, bool)
        self.assertEqual(Annotated.Options[1].annot, float)
        # No type specified annotated, and no default value
        self.assertEqual(Annotated.Options[0].annot, rai.Empty)


if __name__ == '__main__':
    unittest.main()




"""
string_import -- given a string specifying module and object, import it.

This code is ~~stolen from~~ heavily inspired by uvicorn.
"""

from typing import Any

import importlib
import raimad as rai

class StringImportError(ImportError):
    """Error type for errors raised by `string_import()`."""

# TODO Any is not appropriate here
def string_import(import_str: str, multiple: bool = False) -> Any:
    """
    Import a compo, specified by a string.

    Parameters
    ----------
    import_str : str
        A string representing a component or an entire module.
        The string is the the format
        `module.module.module.[...].compo` to import a specific
        compo,
        or 
        `module.module.module.[...]` to import all compos
        in an entire module

    TODO finish this docstring

    """
    module_str, _, attr_str = import_str.partition(':')

    try:
        module = importlib.import_module(module_str)
    except ImportError as exc:
        if exc.name != module_str:
            raise exc from None
        raise StringImportError(
            f'Could not import module "{module_str}".'
            )

    if attr_str == '':

        names = []
        compos = []
        for name, attr in module.__dict__.items():
            if rai.is_compo_class(attr):
                names.append(name)
                compos.append(attr)

        if not compos:
            raise StringImportError(
                f"Module {module} identified by `{module_str}` "
                "does not have any raimad components inside of it."
                )

        elif multiple:
            return compos

        elif len(compos) == 1:
            return compos[0]

        else:
            raise StringImportError(
                f"Module {module} identified by `{module_str}` "
                "has multiple components inside it: "
                f"`{names}`. "
                "You need to tell me which one you want."
                )

    else:
        try:
            attr = getattr(module, attr_str)

        except AttributeError:
            raise StringImportError(
                f"Module {module} identified by `{module_str}` "
                f"does not have an object called `{attr_str}`."
                )

    if not rai.is_compo_class(attr):
        raise StringImportError(
            f"Object `{attr}` of "
            f"module {module} identified by `{module_str}` "
            "is not a raimad component."
            )

    return attr




"""graphviz.py: utilities for generating DOT code."""

from typing import Sequence, Any, TypeAlias, Iterator

class DOTString(str):
    """
    A string containing DOT code.

    This is just a subclass of string that defines a `_repr_dot_`
    method returning self,
    which tells RAIMARK to render it as a DOT picture.
    """

    def _repr_dot_(self) -> str:
        """
        Return self.

        The presence of this method tells RAIMARK to render
        this string as a DOT picture.
        """
        return self


AdjList: TypeAlias = Sequence[tuple[Any, Any]]

def _adjlist2dot(adjlist: AdjList) -> Iterator[str]:
    """Convert adjecency list to DOT code (generator)."""
    yield 'digraph D {\n'
    for from_, to in adjlist:
        yield f'\t {from_} -> {to};\n'
    yield '}\n'

def adjlist2dot(adjlist: AdjList) -> str:
    """Convert adjecency list to DOT code."""
    return ''.join(_adjlist2dot(adjlist))

def _make_dot(adjlist: AdjList, names: dict[int, str]) -> Iterator[str]:
    yield 'digraph D {\n'

    for rout_num, name in names.items():
        yield f'\t {rout_num} [label="{name}"];'

    for from_, to in adjlist:
        yield f'\t {from_} -> {to};\n'

    yield '}\n'

def make_dot(adjlist: AdjList, names: dict[int, str]) -> str:
    """Given an adjacency list and node names, generate DOT code."""
    return ''.join(_make_dot(adjlist, names))




"""affine.py: operations on affine matrices and other math helpers."""

from math import sin, cos, sqrt, atan2
from typing import Sequence

import raimad as rai

def identity() -> 'rai.typing.Affine':
    """Make 3x3 identity matrix."""
    return (
        (1, 0, 0),
        (0, 1, 0),
        (0, 0, 1)
        )

def matmul(*arrays: 'rai.typing.Affine') -> 'rai.typing.Affine':
    """
    Multiply 3x3 matrices.

    Parameters
    ----------
    *arrays: rai.typing.Affine
        The matrices to multiply.
        If only one matrix is passed, that matrix is returned unchanged.
        Otherwise, the product of all the passed matrices is returned.

    Returns
    -------
    rai.typing.Affine
        The product of the matrices.
    """
    if len(arrays) == 1:
        return arrays[0]

    if len(arrays) == 2:
        a, b = arrays
        # This is faster than both numpy and comprehensions,
        # see benchamrks/matmul.py
        # Also, not using numpy has the benefit of not needing a PhD
        # in type gymnastics to get mypy to stop complaining
        return (
            (
                a[0][0] * b[0][0] + a[0][1] * b[1][0] + a[0][2] * b[2][0],
                a[0][0] * b[0][1] + a[0][1] * b[1][1] + a[0][2] * b[2][1],
                a[0][0] * b[0][2] + a[0][1] * b[1][2] + a[0][2] * b[2][2]
                ),
            (
                a[1][0] * b[0][0] + a[1][1] * b[1][0] + a[1][2] * b[2][0],
                a[1][0] * b[0][1] + a[1][1] * b[1][1] + a[1][2] * b[2][1],
                a[1][0] * b[0][2] + a[1][1] * b[1][2] + a[1][2] * b[2][2]
                ),
            (
                a[2][0] * b[0][0] + a[2][1] * b[1][0] + a[2][2] * b[2][0],
                a[2][0] * b[0][1] + a[2][1] * b[1][1] + a[2][2] * b[2][1],
                a[2][0] * b[0][2] + a[2][1] * b[1][2] + a[2][2] * b[2][2]
                )
            )

    return matmul(matmul(arrays[0], arrays[1]), *arrays[2:])

def norm(vec: Sequence[float]) -> float:
    """
    Calculate the norm of a vector (aka equclidean distance).

    Parameters
    ----------
    vec: Sequence[float]
        The vector to compute the norm of.
        It can have any dimension.
        Zero-dimensional vectors are treated as having a norm of zero.

    Returns
    -------
    float
        The norm of the vector.
    """
    # Hardcoding the special cases for vectors of length 0, 1, 2 and 3
    # seems to save a bit of time,
    # see benchmarks/norm.py
    if len(vec) == 0:
        return 0

    if len(vec) == 1:
        return vec[0]

    if len(vec) == 2:
        # `sqrt` is the tiniest bit faster than `** (1/2)`,
        # at least on my machine
        return sqrt(vec[0] ** 2 + vec[1] ** 2)

    if len(vec) == 3:
        return sqrt(vec[0] ** 2 + vec[1] ** 2 + vec[2] ** 2)

    # This is around three times faster than
    # numpy.linalg.norm for small vectors
    return sqrt(sum((coord ** 2 for coord in vec)))


def rotate(angle: float) -> 'rai.typing.Affine':
    """Generate an affine matrix corresponding to a rotation."""
    return (
        (cos(angle), -sin(angle), 0),
        (sin(angle), cos(angle), 0),
        (0, 0, 1),
        )

def move(x: float, y: float) -> 'rai.typing.Affine':
    """Generate an affine matrix corresponding to a translation."""
    return (
        (1, 0, x),
        (0, 1, y),
        (0, 0, 1),
        )

def scale(x: float, y: float) -> 'rai.typing.Affine':
    """Generate an affine matrix corresponding to a scale."""
    return (
        (x, 0, 0),
        (0, y, 0),
        (0, 0, 1),
        )

def around(
        matrix: 'rai.typing.Affine',
        x: float,
        y: float
        ) -> 'rai.typing.Affine':
    """
    Change the origin of an affine matrix.

    This matrix takes an affine matrix and a new origin point.
    It returns a new affine matrix that corresponds to a product of:
    - A translation matrix that puts the new origin into the origin
    - The passed matrix
    - A translation that restores the original origin

    This can be used to perform transformations such as scale or rotation
    around a specific point.
    """
    to_origin = move(-x, -y)
    from_origin = move(x, y)

    return matmul(from_origin, matrix, to_origin)

def get_translation(matrix: 'rai.typing.Affine') -> tuple[float, float]:
    """Given an affine matrix, return the corresponding translation."""
    return matrix[0][2], matrix[1][2]

def get_scale(matrix: 'rai.typing.Affine') -> tuple[float, float]:
    """Given an affine matrix, return the corresponding scale."""
    scale_x = norm((matrix[0][0], matrix[1][0], matrix[2][0]))
    scale_y = norm((matrix[0][1], matrix[1][1], matrix[2][1]))
    return float(scale_x), float(scale_y)

def get_shear(matrix: 'rai.typing.Affine') -> float:
    """Given an affine matrix, return the corresponding shear."""
    scale_x, scale_y = get_scale(matrix)
    return (
        matrix[0][0] * matrix[0][1]
        + matrix[1][0] * matrix[1][1]
        + matrix[2][0] * matrix[2][1]
        ) / (scale_x * scale_y)

def get_rotation(matrix: 'rai.typing.Affine') -> float:
    """Given an affine matrix, return the corresponding rotation."""
    return float(atan2(matrix[1][0], matrix[0][0]))

def transform_xyarray(
        matrix: 'rai.typing.Affine',
        xyarray: 'rai.typing.Poly'
        ) -> 'rai.typing.Poly':
    """Apply transformation to xyarray and return new transformed xyarray."""
    return [
        transform_point(matrix, point)
        for point in xyarray
        ]
    # TODO remove "xyarray name"

def transform_point(
        matrix: 'rai.typing.Affine',
        point: 'rai.typing.PointLike'
        ) -> 'rai.typing.Point':
    """Apply transformation to point and return transformed point."""
    # code written by perplexity.ai
    x, y = point
    a, b, c = matrix[0]
    d, e, f = matrix[1]

    # Apply the transformation
    new_x = a * x + b * y + c
    new_y = d * x + e * y + f

    return (new_x, new_y)




"""layer.py: home to the Layer class."""
import raimad as rai

class Layer(rai.Annotation):
    """
    Layer annotation.

    Layer objects are meant to exist in the `Layers` nested
    class of compos.
    They act as annotations that better explain what each
    layer is.
    """

    def __init__(self, desc: str):
        self.desc = desc
        self.name = 'this should be set in compo.__init_subclass_'




"""iters.py: iteration-related helpers."""

from typing import TypeVar, Any
from collections.abc import Iterable, Callable, Sequence
from itertools import chain


T = TypeVar('T')
def overlap(n: int, seq: Sequence[T]) -> Iterable[Iterable[T]]:
    """
    Iterate n items at a time, with overlap.

    overlap(3, [1,2,3,4,5]) = [
        [1,2,3],
        [2,3,4],
        [3,4,5],
        ]
    """
    return zip(*[seq[offset:] for offset in range(n)])


def nonoverlap(n: int, seq: Sequence[T]) -> Iterable[Iterable[T]]:
    """
    Iterate n items at a time, without overlap.

    Truncates the iterable such that it is a multiple of n.

    overlap(2, [1,2,3,4,5,6]) = [
        [1,2],
        [3,4],
        [5,6],
        ]

    overlap(2, [1,2,3,4,5]) = [
        [1,2],
        [3,4],
        ]
    """
    return zip(*[seq[offset::n] for offset in range(n)])


def _make_alias(
        name: str,
        iterator: Callable[[int, Sequence[T]], Iterable[Iterable[T]]],
        n: int) -> Callable[[Sequence[T]], Iterable[Iterable[T]]]:

    def iterator_alias(seq: Sequence[T]) -> Iterable[Iterable[T]]:
        return iterator(n, seq)

    iterator_alias.__doc__ = (
        f'Iterator through a list {n} items at a time, '
        f'with{"out" if iterator is nonoverlap else ""} overlap.\n'
        f'Equivalent to {iterator.__name__}({n}, iterable)'
        )
    iterator_alias.__name__ = name

    return iterator_alias


duplets = _make_alias('duplets', overlap, 2)
triplets = _make_alias('triplets', overlap, 3)
quadlets = _make_alias('quadlets', overlap, 4)
quintlets = _make_alias('quintlets', overlap, 5)

couples = _make_alias('couples', nonoverlap, 2)
triples = _make_alias('triples', nonoverlap, 3)
quadles = _make_alias('quadles', nonoverlap, 4)
quintles = _make_alias('quintles', nonoverlap, 5)

def flatten(iterable: Iterable[Any]) -> Iterable[Any]:
    """Recursively flatten a nested iterable."""
    if not isinstance(iterable, Iterable) or isinstance(iterable, str):
        # This terminates the recursion
        return [iterable]

    return [
        item
        for sub in iterable
        for item in flatten(sub)
        ]

def braid(*iterables: Iterable[T]) -> Iterable[T]:
    """
    Honestly no clue what this does.

    TODO document and test.
    """
    return list(chain(*zip(*iterables)))

def is_distinct(iterable: Iterable[Any]) -> bool:
    """Return true if no two items in `iterable` are the same."""
    raise NotImplementedError()
    return len(iterable) == len(set(iterable))

def is_rotated(
        first: Sequence[T],
        second: Sequence[T],
        comparison: Callable[[Sequence[T], Sequence[T]], bool] =
                lambda a, b: a == b) -> bool:
    """
    Given two iterables, figure out whether they are "rotated" versions of each other.

    Type of the iterables is not taken into account.

    'abcd', 'bcda' -> True
    [1, 2, 3, 4], (3, 4, 1, 2) -> True
    'abcd', 'bacd' -> False
    'abcd', 'abcde' -> False

    :first: the first iterable
    :second: the other iterable
    :returns: Whether or not one can be obtained by rotating the other
    """
    if len(first) != len(second):
        return False

    length = len(first)
    first_as_tuple = tuple(first)
    # Cast `first` as tuple, since `second` will
    # also be cast to tuple

    for x in range(0, length):
        rotated = (*second[x:], *second[:x])
        assert len(rotated) == length
        print(rotated)
        if comparison(first_as_tuple, rotated):
            return True
    return False





"""Collection of one-size-fits-all wisdoms and questionable quotations."""

fortunes = [

    # Technology #

    'The plural of Regex is "Regrets". ',

    'Why make it simple and efficient, '
    'when you can make it a huge overcomplicated mess? ',

    '''"Stable"? That's for horses. '''

    'Roses are red, violets are blue. '
    'Unterminated string literal at line 32. ',

    'A computer is much like air conditioning: '
    'useless the moment you open Windows. ',

    'We built our world without walls and fences, '
    "so naturally we don't need Windows and Gates. ",

    "object: v. to feel strong distaste for; to oppose firmly.\n"
    "See also: object-oriented programming. ",

    "Just a few hours of trial and error can save "
    "countless minutes of reading the documentation. ",

    "Printing is not supported on this printer. ",

    "Required qualifications: minimal Java experience (ideally none)",

    "Configuring login daemon...\n"
    "Starting login daemon...\n"
    "Exorcising login daemon...\n"
    "System operational.",

    "This is a security feature. It secures our revenue stream. ",

    "We need these permissions to enable intergation. "
    "Integration of our backdoor into your OS, that is. ",

    "Our platform is end-to-end encrypted. "
    "The one end is your client, and the other end is our server. ",

    "We have two-factor authentication. "
    "The first factor is your username, and second factor is your password. ",

    "Unix: full of Daemons, ELFs and DWARFs",

    "Step one: solve the problem. Step two: become the problem. "

    # Education #

    '"The doors are wide open, '
    'but we still see people crashing into the doorframe.\n" '
    '    -- NvdM on IC exams',

    'In order to save energy and help meet our goal of a carbon-neutral '
    'campus by 2025, the light at the end of the tunnel has been turned off. ',

    "Technical university: technically, it's a university. ",

    "Theoretical physicists: in theory, they're good at physics. ",

    "Don't try; just do. ",

    "If you can't, then don't. ",

    "Pay attention, I know exactly what I am talking about. ",

    "Just because [pipelining] wasn't covered in the lectures "
    "doesn't meant it won't be on the exam. ",

    "Learning by doing: I do, you learn. ",

    "There are no stupid questions, "
    "only stupid people. ",

    # Politics #

    "Abuse of authority comes as no surprise. ",

    "Money doesn't bring happiness, but neither does poverty. ",

    '"Somehow the potential to make a lot of money has become more valuable '
    'then making a lot of money lol. " -- u/h3J1e on shareholders',

    "Elect a clown, expect a circus. ",

    "Points of discussion: "
    "building a new shed, and building communism.\n"
    "Due to lumber shortages, it was decided to skip "
    "straight to the second point. ",

    'Now is the time to save face. We hope there will be time to '
    'save other parts of the body as well. ',

    "We wanted the best, but it turned out like always. ",

    "No matter what we try to develop, we always end up with either "
    "the communist party or the Khalashnikov rifle.",

    "You can't hug a child with nuclear arms. ",

    'Experimental missile nicknamed "Civil Servant": '
    "won't work and they can't fire it",

    "We have thoroughly addressed all concerns, from A to B. ",

    "He is a man of logic and should not be taken seriously. ",

    "That is my opinion. Do not ask me why. ",

    "- A choice of what? \n"
    "- A choice of watt? \n"
    "- Yes, a choice of what? \n"
    "- No, the watts are exactly the same, "
    "the choice is who you buy them from. ",

    "- How long are you in for?\n"
    "- Three years\n"
    "- What for?\n"
    "- For nothing!\n"
    "- Do not try to play us for fools, "
    'everyone knows that "for nothing" is 5 years. ',

    "Checklist for travel to The Netherlands:\n"
    "    - Do not travel to The Netherlands. ",

    "Freedom of speech is ofcourse guaranteed. "
    "Freedom *after* the speech, however, is a more delicate topic. ",

    'When I commit violence, it is called "crime", '
    'but when you commit violence, it is called "law". '
    "That doesn't seem fair at all! ",

    "Our enemy is weak and pathetic, but also a deadly threat. ",

    "We pretend to work, and they pretend to pay us. ",

    "The advent of communism will be announced "
    "in the newspapers, over the radio, and perhaps even "
    "on television -- provided the people can still "
    "afford television sets, that is.",

    # Malapropisms #

    "Whatever bloats your goat! ",

    # Engineering #

    "Instructions for use: do not. ",

    "Square peg. Round hole. Get it done. ",

    "Emergency phone not installed. "
    "Please do not have an emergency at this location. ",

    'The difference between practice and theory '
    'is often bigger in practice than in theory. ',

    'No keyboard detected. Press F12 to continue.',

    "Forbidden to remove boulder. "
    "Without it, lift does not carry children. ",

    'Anything labeled "this end towards enemy" is dangerous at both ends. ',

    "It works all the time 60% of the time. ",

    "- I'm from quality assurance. "
    "I can assure you we didn't find any in this product.\n"
    "- Any what?\n"
    "- Any quality.\n",

    "Battery life: up to a minimum of 8 hours! ",

    "This seems like a bad idea, but it's fine for now. ",

    "This was a problem for future me, and now I am future me. ",

    # Resilience #

    "Even with everyone's help, you still have to do everything on your own. ",

    "YOU GOTTA BLAME SOMEONE, RIGHT? BLAME YOURSELF",

    "This time it's the same, but also a little bit different. ",

    "It's not a superiority complex if you're actually superior. ",

    "It's not paranoia if someone's actually out to get you. ",

    "The horrors persist, but so do I",

    "New liver, same eagles. ",

    "It is what it is. ",

    "Life is a joke; don't be the punchline. ",

    "Just because you can survive in harsh conditions "
    "doesn't mean you should have to. ",

    "Even if you're a freak of nature, it doesn't mean that you deserve "
    "to be treated like one. ",

    "Tip: spice up your panic attack with a harmonica. ",

    "Don't delay, despair today! ",

    '''Pessimism: "It's only gonna get worse"\n'''
    '''Optimism: "It *can't* get any worse"''',

    "Cheer up! Things are getting worse at a slower rate! ",

    "Everything is going worryingly well. ",

    "It's easy to miss, but once you see it, you cannot unsee it. ",

    "You can't solve problems with the same "
    "type of thinking that created them. ",

    "This is informative and unfortunate. ",

    '''You're not being "represented", you're being marketed to. ''',

    "You have to laugh about these kinds of things, "
    "because otherwise you will cry. ",

    "Why fall in love when you can just fall asleep? ",

    "Why pass your exams when you can just pass away? ",

    # Misc #

    "The best storytellers are those who have a story to tell. ",

    "CAUTION: Gate swings both ways HARD",

    "On one hand, you've got five oranges. On the other hand, you've got "
    "three apples. Conclusion: you have big hands. ",

    "When you are dead, you do not suffer.\n"
    "The pain is only felt by those around you.\n"
    "The same happens when you are stupid. ",

    "He's suffered for his art and now it's our turn. ",

    "i may be cringe but youre mean and thats worse",

    "Qaulifications: former child",

    '"Whenever a door closes, another one surely opens"\n'
    '    -- My uncle (wise man, but a terrible cabinet maker)',

    "Present day, present time. ",

    "As above, so below. ",

    "I am not insulting you, I am *describing* you. ",

    "I hope this email finds you before I do. ",

    "To whom it WILL concern: ",

    "Bipartisan approval of medical cannabis for arthritis patients:\n"
    "joint support for joint support for joint support. ",

    "The bar is so low that it is practically on the ground, "
    "yet here you are, shovels in hand. ",

    "Every triangle is a love triangle if you love triangles\n"
    " -- Pythagoras",

    "On the fifth day, you gave earth its inhabitants. "
    "You made them in your image, which was probably a mistake. ",

    "What part of hunting/gathering is this supposed to be!?",

    "- I'm telling the truth, I swear! You can just read my mind, can't you?\n"
    "- Heh. Only an idiot would think the truth on Kin Dza Dza!",

    "The birds are singing, the sun is shining, my koi fish pond no longer "
    "has algae in it, and I didn't need to use a single chemical, "
    "and thermaltake is screwing over their customers, "
    "and we're gonna make sure they never get a sale again. ",

    'This is an "A Condition" item. '
    "That is to say, it's definitely in a condition. ",

    "I know writers who use subtext, and they're all cowards.",

    "Two-word horror: Medical Entrepreneur",

    ]




"""compo.py: home of Compo class and supporting constructs."""
import inspect
from typing import Any, NoReturn, Iterator, TypeVar

try:
    from typing import Self
except ImportError:
    # py3.10 and lower
    from typing_extensions import Self

import raimad as rai

class InvalidSubcompoError(TypeError):
    """
    Error for when you try to add something weird as a subcompo.

    Subcompos must be proxies, nothing else.
    """

class ProxyCompoConfusionError(TypeError):
    """(abstract) error for performing operations on compo instead of proxy."""

class CompoInsteadOfProxyAsSubcompoError(
        InvalidSubcompoError,
        ProxyCompoConfusionError
        ):
    """
    Error for when you try to add a Compo instead of a Proxy as a subcompo.

    TODO example
    """

class TransformCompoError(ProxyCompoConfusionError):
    """
    Error for when you try to use transformation function on a compo.

    Compos are supposed to be immutable, you should transform
    a proxy pointing to the compo instead.
    """

class CopyCompoError(ProxyCompoConfusionError):
    """
    Error for when you try to copy a compo instead of its proxy.

    Compos are supposed to be fungible, you should probably
    be copying a proxy pointing to the compo instead.
    """


T = TypeVar('T')
class ProxyableDictList(rai.DictList[T]):
    """TODO document this."""

    _proxy: 'None | rai.typing.Proxy'

    def _post_init(self) -> None:
        self._proxy = None

    def _get_proxy_view(self, proxy: 'rai.t.Proxy') -> Self:
        """TODO document this."""
        new = type(self)(self._dict, copy=False)
        new._proxy = proxy
        return new


class MarksContainer(
        rai.FilteredDictList[
            'rai.typing.Point',
            'rai.typing.PointLike',
            'rai.typing.Point',
            ]):
    """
    Specialized container that implements Compo.marks.

    TODO explain.
    """

    def _filter_set(self, val: 'rai.typing.PointLike') -> 'rai.typing.Point':
        """
        Set filter for markscontainer.

        Make sure that no matter what creature gets passed in
        (regular tuple or boundpoint or whatever),
        what gets stored is a simple regular tuple.
        """
        return (val[0], val[1])

    def _filter_get(self, val: 'rai.typing.Point') -> 'rai.typing.Point':
        return val

class SubcompoContainer(ProxyableDictList['rai.typing.Proxy']):
    """
    Specialized container that implements Compo.subcompos.

    TODO explain.
    """

    def _filter_set(self, item: 'rai.typing.Proxy') -> 'rai.typing.Proxy':
        if isinstance(item, rai.Compo):
            raise CompoInsteadOfProxyAsSubcompoError(
                f"Tried to add object `{item}` of type `{type(item)}` "
                f"as a subcompo of `{self}`. "
                "All subcompos must be of type `raimad.Proxy`."
                "You can call the `.proxy()` method of a `raimad.Compo` "
                "in order to get a proxy that points to it."
                )

        if not isinstance(item, rai.Proxy):
            raise InvalidSubcompoError(
                f"Tried to add object `{item}` of type `{type(item)}` "
                f"as a subcompo of `{self}`. "
                "All subcompos must be of type `raimad.Proxy`."
                )

        return item

    def _filter_get(self, val: 'rai.typing.Proxy') -> 'rai.typing.Proxy':
        if self._proxy is not None:
            return self._proxy.copy_reassign(val, _autogen=True)
        return val

class Compo:
    """
    Compos: the building blocks of RAIMAD.

    TODO explain
    """

    geoms: 'rai.typing.Geoms'
    marks: MarksContainer
    subcompos: SubcompoContainer

    Marks: rai.DictList[rai.Mark]
    Layers: rai.DictList[rai.Layer]
    Options: rai.DictList[rai.Option]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.geoms = {}
        self.subcompos = SubcompoContainer()
        self.marks = MarksContainer()

        self._make(*args, **kwargs)

    def _make(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError()

    @classmethod
    def partial(cls, **kwargs: Any) -> 'rai.typing.Partial':
        """
        Return a Partial of this component class.

        Parameters
        ----------
        kwargs: Any
            Options to store in the Partial

        Returns
        -------
        rai.typing.Partial
            The new Partial
        """
        return rai.Partial(cls, **kwargs)

    def steamroll(self) -> 'rai.typing.Geoms':
        """
        Steamroll the entire compo hierarchy into one Geoms dict.

        TODO more informative
        """
        geoms = self.geoms.copy()
        for subcompo in self.subcompos.values():
            # TODO override "update" method in geoms container?
            for layer_name, layer_geoms in subcompo.steamroll().items():
                if layer_name not in geoms.keys():
                    geoms[layer_name] = []
                geoms[layer_name].extend(layer_geoms)
        return geoms

    def final(self) -> Self:
        """
        Return self.

        This method exists for uniformity with Proxy.final
        """
        return self

    def depth(self) -> int:
        """
        Return the number zero.

        This method exists for uniformity with Proxy.depth
        """
        return 0

    def descend(self) -> Iterator[Self]:
        """
        Yield self.

        This method exists for uniformity with Proxy.descend
        """
        yield self

    def descend_p(self) -> 'Iterator[rai.typing.Proxy]':
        """
        Yield nothing.

        This method exists for uniformity with Proxy.descend_p
        """
        return
        yield

    def proxy(self) -> 'rai.typing.Proxy':
        """Return new Proxy pointing to this Compo."""
        return rai.Proxy(self)

    @property
    def copy(self) -> NoReturn:
        """Deliberately unimplemented -- see Proxy.copy()."""
        raise CopyCompoError(
            f"`{self}` is a Compo, not a Proxy! "
            "Don't copy compos; instead, "
            "create a Proxy using the `.proxy()` method "
            "and copy that instead."
            )

    def walk_hier(self) -> Iterator['rai.typing.CompoLike']:
        """
        Traverse the subcomponent hierarchy of this compo.

        This method will recursively walk through the entire
        subcompo hierarchy of this compo, including self.
        """
        yield self
        for subcompo in self.subcompos.values():
            yield from subcompo.walk_hier()

    def transform_point(
            self,
            point: 'rai.typing.PointLike'
            ) -> 'rai.typing.PointLike':
        """
        Do nothing to `point` and return as-is.

        This method exists for uniformity with Proxy.transform_point.
        """
        return point

    @property
    def scale(self) -> NoReturn:
        """Deliberately unimplemented -- see Proxy.scale."""
        raise TransformCompoError(
            f"Tried to scale `{self}`, which is a Compo. "
            "Compos are not transformable; call the `.proxy()` method "
            "to get a Proxy pointing to this compo and transform that instead."
            )

    @property
    def move(self) -> NoReturn:
        """Deliberately unimplemented -- see Proxy.move."""
        raise TransformCompoError(
            f"Tried to move `{self}`, which is a Compo. "
            "Compos are not transformable; call the `.proxy()` method "
            "to get a Proxy pointing to this compo and transform that instead."
            )

    @property
    def movex(self) -> NoReturn:
        """Deliberately unimplemented -- see Proxy.movex."""
        raise TransformCompoError(
            f"Tried to move `{self}`, which is a Compo. "
            "Compos are not transformable; call the `.proxy()` method "
            "to get a Proxy pointing to this compo and transform that instead."
            )

    @property
    def movey(self) -> NoReturn:
        """Deliberately unimplemented -- see Proxy.movey."""
        raise TransformCompoError(
            f"Tried to move `{self}`, which is a Compo. "
            "Compos are not transformable; call the `.proxy()` method "
            "to get a Proxy pointing to this compo and transform that instead."
            )

    @property
    def rotate(self) -> NoReturn:
        """Deliberately unimplemented -- see Proxy.rotate."""
        raise TransformCompoError(
            f"Tried to rotate `{self}`, which is a Compo. "
            "Compos are not transformable; call the `.proxy()` method "
            "to get a Proxy pointing to this compo and transform that instead."
            )

    @property
    def flip(self) -> NoReturn:
        """Deliberately unimplemented -- see Proxy.flip."""
        raise TransformCompoError(
            f"Tried to flip `{self}`, which is a Compo. "
            "Compos are not transformable; call the `.proxy()` method "
            "to get a Proxy pointing to this compo and transform that instead."
            )

    @property
    def hflip(self) -> NoReturn:
        """Deliberately unimplemented -- see Proxy.hflip."""
        raise TransformCompoError(
            f"Tried to hflip `{self}`, which is a Compo. "
            "Compos are not transformable; call the `.proxy()` method "
            "to get a Proxy pointing to this compo and transform that instead."
            )

    @property
    def vflip(self) -> NoReturn:
        """Deliberately unimplemented -- see Proxy.vflip."""
        raise TransformCompoError(
            f"Tried to vflip `{self}`, which is a Compo. "
            "Compos are not transformable; call the `.proxy()` method "
            "to get a Proxy pointing to this compo and transform that instead."
            )
        # TODO tests to make sure the transform functions are all the same
        # in transform, proxy, and compo

    # bbox functions #
    @property
    def bbox(self) -> 'rai.BBox':
        """Get bbox of this compo."""
        bbox = rai.BBox()
        for geoms in self.steamroll().values():
            for geom in geoms:
                bbox.add_xyarray(geom)
        return bbox

    def __init_subclass__(cls) -> None:
        """Bookkeeping for creating new Compo classes."""
        _class_to_dictlist(cls, 'Marks', rai.Mark)
        _class_to_dictlist(cls, 'Layers', rai.Layer)
        _class_to_dictlist(cls, 'Options', rai.Option)

        for param in inspect.signature(cls._make).parameters.values():
            if param.name not in cls.Options.keys():
                # TODO unannotated
                continue

            cls.Options[param.name].annot = rai.Empty
            cls.Options[param.name].default = rai.Empty

            if param.default is not inspect._empty:
                cls.Options[param.name].default = param.default

            if param.annotation is inspect._empty:
                if param.default is not inspect._empty:
                    cls.Options[param.name].annot = type(param.default)
            else:
                cls.Options[param.name].annot = param.annotation

    def auto_subcompos(self, locs: dict[Any, Any] | None = None) -> None:
        """
        Dirty hack for use in `_make` to quickly add all proxies as subcompos.

        Simply call `self.auto_subcompos(locals())` at the
        end of your `_make()` function, and all of the
        proxies in the scope of `_make()` will be added as
        subcompos.
        If you do no provide `locals()`, this function
        will instead use arcane `inspect` magic
        to traverse the stack and extract them automatically.

        ...or, better yet, do not use this function and
        add subcompos explicitly.
        """
        locs = locs or inspect.stack()[1].frame.f_locals

        for name, obj in locs.items():
            if (
                    isinstance(obj, rai.Proxy)
                    and name != 'self'
                    and not name.startswith('_')
                    ):
                # TODO forbid adding compo directly as subcompo,
                # only proxy, instead of doing it automatically.
                self.subcompos[name] = obj

    def __str__(self) -> str:
        """Get string representation of compo."""
        return (
            "<"
            f"{type(self).__name__} at {rai.wingdingify(id(self))} "
            ">"
            )

    def __repr__(self) -> str:
        """Get string representation of compo."""
        return self.__str__()

    def _repr_svg_(self) -> str:
        """
        Make svg representation of component.

        This is called by jupyter and raimark.
        """
        return rai.export_svg(self)


K = TypeVar('K', bound=rai.Annotation)
def _class_to_dictlist(
        cls: type,
        attr: str,
        wanted_type: type[K]
        ) -> None:

    if not hasattr(cls, attr):
        setattr(cls, attr, rai.DictList())
        return

    new_list: rai.DictList[K] = rai.DictList()
    for name, annot in getattr(cls, attr).__dict__.items():
        if not isinstance(annot, wanted_type):
            continue
        annot.name = name
        new_list[name] = annot

    setattr(cls, attr, new_list)




import unittest
from math import radians

import raimad as rai

from .utils import ArrayAlmostEqual

class TwoCircles(rai.Compo):
    """
    TwoCircles
    Two circles of radius 5 side-by-side on the x axis
    """
    def _make(self):
        self.subcompos.append(
            rai.Circle(5).proxy().movex(-5)
            )
        self.subcompos.append(
            rai.Circle(5).proxy().movex(5)
            )

class RotatedCircles(rai.Compo):
    """
    RotatedCircles
    Two circles of radius 5 side-by-side on the y axis
    """
    def _make(self):
        self.subcompos.append(
            TwoCircles().proxy().rotate(radians(90))
            )

class TestBBox(ArrayAlmostEqual, unittest.TestCase, epsilon=0.01):

    def test_bbox_basic(self):
        circle = rai.Circle(5)

        self.assertArrayAlmostEqual(
            circle.bbox,
            [-5, -5, 5, 5]
            )
        self.assertAlmostEqual(circle.bbox.length, 10)
        self.assertAlmostEqual(circle.bbox.width, 10)
        self.assertAlmostEqual(circle.bbox.top, 5)
        self.assertAlmostEqual(circle.bbox.bottom, -5)
        self.assertAlmostEqual(circle.bbox.left, -5)
        self.assertAlmostEqual(circle.bbox.right, 5)
        self.assertArrayAlmostEqual(
            circle.bbox.top_left,
            (-5, 5)
            )
        self.assertArrayAlmostEqual(
            circle.bbox.top_mid,
            (0, 5)
            )
        self.assertArrayAlmostEqual(
            circle.bbox.top_right,
            (5, 5)
            )
        self.assertArrayAlmostEqual(
            circle.bbox.mid_left,
            (-5, 0)
            )
        self.assertArrayAlmostEqual(
            circle.bbox.mid,
            (0, 0)
            )
        self.assertArrayAlmostEqual(
            circle.bbox.mid_right,
            (5, 0)
            )
        self.assertArrayAlmostEqual(
            circle.bbox.bot_left,
            (-5, -5)
            )
        self.assertArrayAlmostEqual(
            circle.bbox.bot_mid,
            (0, -5)
            )
        self.assertArrayAlmostEqual(
            circle.bbox.bot_right,
            (5, -5)
            )

    def test_bbox_subcompo(self):
        circles = TwoCircles()

        self.assertArrayAlmostEqual(
            circles.bbox,
            [-10, -5, 10, 5]
            )
        self.assertAlmostEqual(circles.bbox.length, 20)
        self.assertAlmostEqual(circles.bbox.width, 10)
        self.assertAlmostEqual(circles.bbox.top, 5)
        self.assertAlmostEqual(circles.bbox.bottom, -5)
        self.assertAlmostEqual(circles.bbox.left, -10)
        self.assertAlmostEqual(circles.bbox.right, 10)
        self.assertArrayAlmostEqual(
            circles.bbox.top_left,
            (-10, 5)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.top_mid,
            (0, 5)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.top_right,
            (10, 5)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.mid_left,
            (-10, 0)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.mid,
            (0, 0)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.mid_right,
            (10, 0)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.bot_left,
            (-10, -5)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.bot_mid,
            (0, -5)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.bot_right,
            (10, -5)
            )

    def test_bbox_transform(self):
        circles = RotatedCircles()

        self.assertArrayAlmostEqual(
            circles.bbox,
            [-5, -10, 5, 10]
            )
        self.assertAlmostEqual(circles.bbox.length, 10)
        self.assertAlmostEqual(circles.bbox.width, 20)
        self.assertAlmostEqual(circles.bbox.top, 10)
        self.assertAlmostEqual(circles.bbox.bottom, -10)
        self.assertAlmostEqual(circles.bbox.left, -5)
        self.assertAlmostEqual(circles.bbox.right, 5)
        self.assertArrayAlmostEqual(
            circles.bbox.top_left,
            (-5, 10)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.top_mid,
            (0, 10)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.top_right,
            (5, 10)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.mid_left,
            (-5, 0)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.mid,
            (0, 0)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.mid_right,
            (5, 0)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.bot_left,
            (-5, -10)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.bot_mid,
            (0, -10)
            )
        self.assertArrayAlmostEqual(
            circles.bbox.bot_right,
            (5, -10)
            )

    def test_bbox_pad(self):
        circles = RotatedCircles()
        bbox = circles.bbox

        self.assertArrayAlmostEqual(
            bbox,
            [-5, -10, 5, 10]
            )

        self.assertArrayAlmostEqual(
            bbox.pad(.1),
            [-5.1, -10.1, 5.1, 10.1]
            )

        self.assertArrayAlmostEqual(
            bbox.pad(.2, .3),
            [-5.2, -10.3, 5.2, 10.3]
            )

        with self.assertRaises(TypeError):
            bbox.pad(.2, .3, .4)

        with self.assertRaises(TypeError):
            bbox.pad(.2, .3, .4, .5)

        self.assertArrayAlmostEqual(
            bbox.pad(left=0.3),
            [-5.3, -10, 5, 10]
            )

        self.assertArrayAlmostEqual(
            bbox.pad(bottom=0.4),
            [-5, -10.4, 5, 10]
            )

        self.assertArrayAlmostEqual(
            bbox.pad(right=0.2, bottom=0.4),
            [-5, -10.4, 5.2, 10]
            )

        self.assertArrayAlmostEqual(
            bbox.pad(right=0.3, bottom=0.4, left=0.1),
            [-5.1, -10.4, 5.3, 10]
            )

        self.assertArrayAlmostEqual(
            bbox.pad(.2, top=0.3, bottom=0.4, left=0.1),
            [-5 - .2 - .1, -10 - .2 - .4, 5 + .2, 10 + .2 + .3]
            )

        self.assertArrayAlmostEqual(
            bbox.pad(.2, .1, top=0.3, bottom=0.4, left=0.1),
            [-5 - .2 - .1, -10 - .1 - .4, 5 + .2, 10 + .1 + .3]
            )

    def test_bbox_manual(self):
        bbox = rai.BBox()

        self.assertTrue(bbox.is_empty())

        with self.assertRaises(rai.err.EmptyBBoxError):
            bbox.assert_nonempty()

        with self.assertRaises(rai.err.EmptyBBoxError):
            bbox.top

        with self.assertRaises(rai.err.EmptyBBoxError):
            bbox.top_mid

        with self.assertRaises(rai.err.EmptyBBoxError):
            bbox.pad(1, 1)

        bbox.add_point((1, 1))

        self.assertArrayAlmostEqual(
            bbox,
            [1, 1, 1, 1]
            )

        self.assertFalse(bbox.is_empty())

        # all of these should NOT raise EmptyBBoxError
        bbox.assert_nonempty()
        bbox.top
        bbox.top_mid
        bbox.pad(1, 1)

        bbox.add_point((-2, -3))

        self.assertArrayAlmostEqual(
            bbox,
            [-2, -3, 1, 1]
            )

    def test_bbox_snapping(self):
        c1 = rai.Proxy(rai.Circle(10))
        c2 = rai.Proxy(rai.Circle(5))
        c3 = rai.Proxy(rai.Circle(2))

        c1.bbox.mid.to((0, 0))
        c2.snap_left(c1)
        c3.bbox.top_mid.to(c1.bbox.mid_right)

        self.assertArrayAlmostEqual(c1.bbox.mid, (0, 0))
        self.assertArrayAlmostEqual(c2.bbox.mid, (-10 - 5, 0))
        self.assertArrayAlmostEqual(c3.bbox.mid, (10, -2))

    def test_bbox_bound(self):
        c1 = rai.Circle(10)
        unbound = c1.bbox

        with self.assertRaises(AttributeError):
            unbound.mid.to((0, 0))

        # TODO there's no easy way here to write an error message like
        # "hey, you tried to do a transform through an unbound bbox,
        # you need to do it through a bound bbox"
        # All you get is a cryptic "tuple has no attribute 'to'"
        # Maybe we need a raidoc page with common errors?

    def test_bbox_copy_transform(self):
        c1 = rai.Circle(10).proxy()

        c1.bbox.pad(5).mid_left.to((0, 0))

        self.assertArrayAlmostEqual(
            c1.bbox.mid,
            (15, 0)
            )


if __name__ == '__main__':
    unittest.main()




"""empty.py: Home to EmptyType and Empty."""

from typing import ClassVar

class EmptyType:
    """
    Empty: an object representing an empty field.

    This can be used similar to `None`,
    except in cases where `None` has a different meaning.
    For example, compo options by default have their
    `browser_default` property set to Empty,
    indicating that there is no default value to use
    in the component browser.
    Setting it to None would mean that the default value is None.

    This is a singleton class, meaning only one Empty exists
    in the entire world.
    If you try to instantiate a new Empty,
    you will just get the same one.
    """

    instance: ClassVar['EmptyType']

    def __new__(cls) -> 'EmptyType':
        """Return the one Empty or create it if missing."""
        if not hasattr(cls, 'instance'):
            cls.instance = super(EmptyType, cls).__new__(cls)
        return cls.instance

    def __bool__(self) -> bool:
        """Get boolean representation of Empty (always False)."""
        return False

    def __str__(self) -> str:
        """Get string representation of Empty."""
        return "<Empty>"

    def __repr__(self) -> str:
        """Get string representation of Empty."""
        return "<Empty>"


Empty: EmptyType = EmptyType()




"""boundpoint.py: home to BoundPoint class."""

from typing import Literal, Iterator

import raimad as rai

class BoundPoint():
    """
    A point bound to a Proxy.

    There is often a desire to perform a certain transformation
    of some proxy in reference to a given point at that proxy.
    We might want to rotate a proxy around the top left corner
    of its bbox,
    or move a proxy such that one of its marks is at a given point.

    These things can be done with the following syntax in raimad:
    - `someproxy.bbox.top_left.rotate(someangle)`
    - `someproxy.marks.important_mark.to((x, y))`

    In the examples above, `top_left` and `important_mark` are both
    BoundPoints.
    They behave just like regular points (a tuple of x and y),
    but they also define transformation methods like
    `rotate`, `to`, `vflip`, and so forth
    that perform the corresponding transformation
    to the bound proxy, with the BoundPoint treated as the new origin
    for the transformation.
    """
    def __init__(self, x: float, y: float, proxy: 'rai.typing.Proxy'):
        self._x = x
        self._y = y
        self._proxy = proxy

    def __getitem__(self, index: Literal[0, 1]) -> float:
        """
        Get the X or Y coordinate of this BoundPoint.

        BoundPoints behave just like regular Points
        (tuple of x and y),
        meaning that you can get the X and Y coordinates using
        the square bracket syntax:
            - `somepoint[0]` <-- x coordinate
            - `somepoint[y]` <-- y coordinate

        Parameters
        ----------
        index: int
            0 or 1 for the x and y coordinate respectively

        Returns
        -------
        float
            The coordinate value

        Raises
        ------
            IndexError is raised if `index` is anything other than 0 or 1.
        """
        if index == 0:
            return self._x

        elif index == 1:
            return self._y

        # The esxception type HAS TO BE IndexError
        # if we want Python to allow unpacking BoundPoints
        # using the asterisk syntax
        raise IndexError(
            "BoundPoint has only x and y coordinates, "
            "so the index must be either 0 or 1"
            )

    def __iter__(self) -> Iterator[float]:
        """Return an iterator of the x and y coordinates of this BoundPoint."""
        return iter((self._x, self._y))

    def __eq__(self, other: object) -> bool:
        """
        Check for equality of BoundPoint.

        Parameters
        ----------
        other: object
            The object to check equality against.
            This may be another BoundPoint, a regular Point,
            or any other object.
            If `other` does not implement a `__getitem__` method,
            then the result is automatically zero.
            If `other` does implement `__getitem__`,
            then the result is True iff
            the x coordinate of this boundpoint == other[0] and
            the y coordinate of this boundpoint == other[1].

        Returns
        -------
        bool
            True if this boundpoint is equal to `other`,
            False otherwise.

        Notes
        -----
        Comparing BoundPoints using the == operator
        (i.e. this method) may be a bad idea for the same reason
        that comparing floats in this way is a bad idea.

        TODO add a `distance_between` helper method?
        """
        if not hasattr(other, "__getitem__"):
            return False

        return bool(
            (self._x == other[0])
            and
            (self._y == other[1])
            )

    def __repr__(self) -> str:
        """Return string representation of this BoundPoint."""
        return f'<({self._x}, {self._y}) bound to {self._proxy}>'

    def to(self, point: 'rai.typing.PointLike') -> 'rai.typing.Proxy':
        """
        Move the bound Proxy such that this BoundPoint aligns with `point`.

        Parameters
        ----------
        point: rai.typing.PointLike
            Target point.

        Returns
        -------
        rai.typing.Proxy
            The bound proxy (not this BoundPoint!) is returned
            to allow chaining methods.
        """
        self._proxy.transform.move(
            point[0] - self._x,
            point[1] - self._y,
            )
        return self._proxy

    def rotate(self, angle: float) -> 'rai.typing.Proxy':
        """
        Rotate the bound Proxy around this BoundPoint.

        Parameters
        ----------
        angle: float
            The rotation angle, specified in radians in the positive
            direction.

        Returns
        -------
        rai.typing.Proxy
            The bound proxy (not this BoundPoint!) is returned
            to allow chaining methods.
        """
        self._proxy.transform.rotate(
            angle,
            self._x,
            self._y
            )
        return self._proxy

    def move(self, x: float, y: float) -> 'rai.typing.Proxy':
        """
        Move the bound Proxy.

        Since translation is a transformation that works the same
        regardless of where the origin is,
        this method is identicaly to simply calling the `move`
        method of the bound transform itself.

        Parameters
        ----------
        x: float
            Move this many units in the X direction
        y: float
            Move this many units in the Y direction

        Returns
        -------
        rai.typing.Proxy
            The bound proxy (not this BoundPoint!) is returned
            to allow chaining methods.
        """
        self._proxy.transform.move(x, y)
        return self._proxy

    def flip(self) -> 'rai.typing.Proxy':
        """
        TODO add tests
        """
        self._proxy.transform.flip(self._x, self._y)
        return self._proxy

    def hflip(self) -> 'rai.typing.Proxy':
        """
        TODO add tests
        """
        self._proxy.transform.flip(self._x)
        return self._proxy

    def vflip(self) -> 'rai.typing.Proxy':
        """
        TODO add tests
        """
        self._proxy.transform.flip(self._y)
        return self._proxy

    # TODO the rest of the functions




import unittest
from math import sin, cos, radians

import raimad as rai

class TestAffine(unittest.TestCase):
    def test_identity(self):
        self.assertEqual(
            rai.affine.identity(),
            ((1, 0, 0), (0, 1, 0), (0, 0, 1))
            )

    def test_matmul_singular(self):
        mat1 = (
            (1, 2, 3),
            (4, 5, 6),
            (7, 8, 9),
            )

        self.assertEqual(
            rai.affine.matmul(mat1),
            mat1
            )

    def test_matmul_double(self):
        mat1 = (
            (1, 2, 3),
            (4, 5, 6),
            (7, 8, 9),
            )

        mat2 = (
            (4, 5, 6),
            (1, 2, 3),
            (7, 8, 9),
            )

        self.assertEqual(
            rai.affine.matmul(mat1, mat2),
            (
                (27,  33,  39),
                (63,  78,  93),
                (99, 123, 147),
                )
            )

    def test_matmul_triple(self):
        mat1 = (
            (1, 2, 3),
            (4, 5, 6),
            (7, 8, 9),
            )

        mat2 = (
            (4, 5, 6),
            (1, 2, 3),
            (7, 8, 9),
            )

        mat3 = (
            (4, 1, 6),
            (1, 1, 1),
            (7, 1, 9),
            )

        self.assertEqual(
            rai.affine.matmul(mat1, mat2, mat3),
            (
                (414, 99, 546),
                (981, 234, 1293),
                (1548, 369, 2040),
                )
            )

    def test_norm_empty(self):
        self.assertEqual(
            rai.affine.norm([]),
            0
            )

    def test_norm_single(self):
        self.assertEqual(
            rai.affine.norm([420]),
            420
            )

    def test_norm_double(self):
        self.assertAlmostEqual(
            rai.affine.norm([3, 4]),
            5
            )

    def test_norm_triple(self):
        self.assertAlmostEqual(
            rai.affine.norm([3, 4, 5]),
            7.0710678118654755
            )

    def test_norm_quad(self):
        self.assertAlmostEqual(
            rai.affine.norm([3, 4, 5, 6]),
            9.273618495495704
            )

    def test_get_matrix_translation(self):
        x, y = rai.affine.get_translation(
            (
                (1, 0, 6),
                (0, 1, 9),
                (0, 0, 1),
                )
            )
        self.assertAlmostEqual(x, 6)
        self.assertAlmostEqual(y, 9)

    def test_get_matrix_scale(self):
        x, y = rai.affine.get_scale(
            (
                (6, 0, 0),
                (0, 9, 0),
                (0, 0, 1),
                )
            )
        self.assertAlmostEqual(x, 6)
        self.assertAlmostEqual(y, 9)

    def test_get_matrix_shear(self):
        shear = rai.affine.get_shear(
            (
                (1, 0.5, 0),
                (0, 1, 0),
                (0, 0, 1),
                )
            )
        self.assertAlmostEqual(shear, 0.4472135954999)

    def test_get_matrix_rotation(self):
        angle = rai.affine.get_rotation(
            (
                (cos(radians(30)), -sin(radians(30)), 0),
                (sin(radians(30)), cos(radians(30)), 0),
                (0, 0, 1),
                )
            )
        self.assertAlmostEqual(angle, radians(30))
        


if __name__ == '__main__':
    unittest.main()




import unittest

import raimad as rai

from .utils import GeomsEqual, ArrayAlmostEqual

class TestPolys(GeomsEqual, ArrayAlmostEqual, unittest.TestCase):

    def test_rectwire_p2p(self):
        rectwire = rai.RectWire.from_points((0, 0), (0, 20), 10)
        self.assertGeomsEqual(
            rectwire.geoms,
            {
                'root': [
                    [
                        [-5, 0],
                        [5, 0],
                        [5, 20],
                        [-5, 20]
                        ],
                    ]
                }
            )

    def test_rectwire_shortsyntax(self):
        rectwire = rai.RectWire((0, 0), (0, 20), 10)
        self.assertGeomsEqual(
            rectwire.geoms,
            {
                'root': [
                    [
                        [-5, 0],
                        [5, 0],
                        [5, 20],
                        [-5, 20]
                        ],
                    ]
                }
            )

    def test_rectwire_polar(self):
        rectwire = rai.RectWire.from_polar(
            (0, 0),
            angle=rai.quartercircle,
            length=20,
            width=10,
            )
        self.assertGeomsEqual(
            rectwire.geoms,
            {
                'root': [
                    [
                        [-5, 0],
                        [5, 0],
                        [5, 20],
                        [-5, 20]
                        ],
                    ]
                }
            )


if __name__ == '__main__':
    unittest.main()




"""
shorthand.py: defines `export_cif` function.

The `export_cif` function is a more convenient way
of exporting components to CIF than using an exporter
directly.
"""

from pathlib import Path
from typing import TextIO, Protocol, Any

import raimad as rai

class InvalidDestinationError(ValueError):
    """Error raised by `export_cif` when `dest` parameter is set incorrectly."""

class ExporterProto(Protocol):
    """
    Protocol for CIF exporters.

    Currently, there is only one CIF exporter in RAIMAD: NoReuse.
    Future exporters (and user-defined exporters) must
    conform to this Protocol.

    If you don't know what Protocols are,
    they're like a "shell" of a class that specifies what features must
    be present in an actual class in order for it to comply
    (in this case, an __init__ method with certain paramters,
    and a `cif_string` attribute).

    Just like all othr type annotation,
    protocols do nothing at runtime.
    They're checked statically by mypy.
    """

    def __init__(
            self,
            compo: 'rai.typing.CompoLike',
            *args: Any,
            **kwargs: Any
            ) -> None:
        ...

    cif_string: str

def export_cif(
        compo: 'rai.typing.CompoLike',
        dest: str | Path | TextIO | None = None,
        exporter: type[ExporterProto] | None = None,
        *args: Any,
        **kwargs: Any,
        ) -> str:
    """
    Export component to CIF.

    Parameters
    ----------
    compo: rai.typing.CompoLike
        The compo or proxy to export

    dest: str | Path | TextIO | None
        A path to a file or an already open file for saving the CIF
        output.
        If, instead of saving to disk, you want to have the CIF code
        as a string, set `dest` to None (or leave unset)
        and use this functions return value.

    exporter: ExporterProto
        The exporter to use. Currently, there is only one CIF exporter
        in RAIMAD.

    *args: Any
        Additional arguments will be passed to the the Exporter's __init__.

    *args: Any
        Additional keyword arguments will be passed
        to the the Exporter's __init__.

    Returns
    -------
        The generated CIF code is always returned as a string,
        regardless of what `dest` is set to.
    """
    exporter_instance = (exporter or rai.cif.NoReuse)(compo, *args, **kwargs)
    cif_string = exporter_instance.cif_string

    if dest is None:
        pass
    elif isinstance(dest, (str, Path)):
        with open(dest, 'w') as file:
            file.write(cif_string)
    elif hasattr(dest, 'write'):
        dest.write(cif_string)
    else:
        raise InvalidDestinationError(
            f"Invalid destination type {type(dest)}. "
            "Must be a file path or a file-like stream."
            )

    return cif_string




import unittest

import raimad as rai

from .utils import ArrayAlmostEqual

class BareGeometric(rai.Compo):
    def _make(self):
        self.geoms.update({
            'root': [
                [
                    [0, 0],
                    [0, 10],
                    [10, 10],
                    [10, 0],
                    ],
                [
                    [20, 20],
                    [40, 20],
                    [30, 40],
                    ],
                ]
            })
        self.marks.triangle_corner = (20, 40)

class Intermid(rai.Compo):
    def _make(self):
        self.subcompos.append(
            rai.Proxy(
                BareGeometric(),
                {'root': 'intermediary'},
                rai.Transform().movex(-3)
                )
            )
        self.marks.propagated = self.subcompos[0].marks.triangle_corner

class BareStructural(rai.Compo):
    def _make(self):
        self.subcompos.append(
            rai.Proxy(
                BareGeometric(),
                {'root': 'upper'},
                rai.Transform().scale(2)
                )
            )
        self.subcompos.append(
            rai.Proxy(
                BareGeometric(),
                {'root': 'lower'},
                rai.Transform().scale(0.5)
                )
            )
        self.subcompos.append(
            rai.Proxy(
                Intermid(),
                {'intermediary': 'lower'},
                rai.Transform().movey(-3)
                )
            )

        self.marks.propagated = self.subcompos[2].marks.propagated

class BareGeometricSyntax(rai.Compo):
    def _make(self):
        self.geoms.update({
            'root': [
                [
                    [0, 0],
                    [0, 10],
                    [10, 10],
                    [10, 0],
                    ],
                [
                    [20, 20],
                    [40, 20],
                    [30, 40],
                    ],
                ]
            })
        self.marks.triangle_corner = (20, 40)

class IntermidSyntax(rai.Compo):
    def _make(self):
        self.subcompos.append(
            rai.Proxy(
                BareGeometricSyntax(),
                {'root': 'intermediary'},
                rai.Transform().movex(-3)
                )
            )
        self.marks.propagated = self.subcompos[0].marks.triangle_corner

class BareStructuralSyntax(rai.Compo):
    def _make(self):
        self.subcompos.append(
            rai.Proxy(
                BareGeometricSyntax(),
                {'root': 'upper'},
                rai.Transform().scale(2)
                )
            )
        self.subcompos.append(
            rai.Proxy(
                BareGeometricSyntax(),
                {'root': 'lower'},
                rai.Transform().scale(0.5)
                )
            )
        self.subcompos.append(
            rai.Proxy(
                IntermidSyntax(),
                {'intermediary': 'lower'},
                rai.Transform().movey(-3)
                )
            )

        self.marks.propagated = self.subcompos[2].marks.propagated

class TestMarks(ArrayAlmostEqual, unittest.TestCase):

    def test_marks(self):
        compo = BareStructural()

        self.assertArrayAlmostEqual(
            compo.marks.propagated,
            (20 - 3, 40 - 3)
            )

        self.assertArrayAlmostEqual(
            compo.subcompos[2].subcompos[0].marks.triangle_corner,
            (20 - 3, 40 - 3)
            )

        self.assertArrayAlmostEqual(
            compo.subcompos[2].marks.propagated,
            (20 - 3, 40 - 3)
            )

    def test_marks_syntax(self):
        compo = BareStructuralSyntax()

        self.assertArrayAlmostEqual(
            compo.marks.propagated,
            (20 - 3, 40 - 3)
            )

        self.assertArrayAlmostEqual(
            compo.subcompos[2].subcompos[0].marks.triangle_corner,
            (20 - 3, 40 - 3)
            )

        self.assertArrayAlmostEqual(
            compo.subcompos[2].marks.propagated,
            (20 - 3, 40 - 3)
            )


if __name__ == '__main__':
    unittest.main()




"""snowman.py: home to Snowman compo."""

import raimad as rai

class Snowman(rai.Compo):
    """A sample component."""

    browser_tags = ["builtin", "example"]

    class Marks:
        nose = rai.Mark("Tip of the snowman's nose")

    class Options:
        nose_length = rai.Option.Geometric("Length of nose")
        eye_size = rai.Option.Geometric("Eye radius")

    class Layers:
        pebble = rai.Layer(
            "Non-malleable polycrystalline silicon layer"
            )
        carrot = rai.Layer(
            "Bio-lithographic layer characterised by lambda = approx. 6E-7nm"
            )
        snow = rai.Layer(
            "A collection of individual crystals of frozen dihydrogen monoxide"
            )

    def _make(
            self,
            nose_length: float = 10,
            eye_size: float = 2,
            ) -> None:

        base = rai.Circle(50).proxy().map('snow')
        torso = rai.Circle(40).proxy().map('snow')
        head = rai.Circle(20).proxy().map('snow')

        torso.snap_above(base)
        head.snap_above(torso)

        eye_l = rai.Circle(eye_size).proxy().map('pebble')

        eye_l.marks.center.to(
            head.bbox.interpolate(0.3, 0.7)
            )

        eye_r = eye_l.copy()
        eye_r.vflip(head.bbox.mid[0])
        # TODO flip at arbitrary angle around point?
        # also, for hflip/vflip, is passed point,
        # automatically pick coordinate.

        nose = rai.CustomPoly([
            (0, 2),
            ('tip', (nose_length, 0)),
            (0, -2)
            ]).proxy().map('carrot')

        nose.move(
            *head.bbox.interpolate(0.5, 0.5)
            )

        self.marks.nose = nose.marks.tip

        self.subcompos.base = base
        self.subcompos.torso = torso
        self.subcompos.head = head
        self.subcompos.eye_l = eye_l
        self.subcompos.eye_r = eye_r
        self.subcompos.nose = nose




"""annotation.py: home to Annotation class."""
import raimad as rai

class Annotation:
    """
    Abstract base class for Options, Layers, and Marks annotations.

    Declares that all of those things have a `name` attribute,
    so mypy does not complain about `compo.py/_class_to_dictlist()`
    """

    name: str | rai.EmptyType




"""partial.py: home to the Partial class."""

from typing import Any
from copy import copy
import raimad as rai

class Partial:
    """
    Partial: a halfway-created Compo.

    Much like Pythons `functools.partial`,
    which allows partially calling a function,
    a Partial allows partially creating a Compo.
    A Partial points to a CompoType and holds a set of
    pre-determined options for that CompoType.
    When the Partial is __call__'ed,
    it creates a new Compo based on the held options,
    as well as any options passed to __call__.
    The options passed to __call__ take precedence
    over the held options.
    """

    compo_cls: 'rai.typing.CompoType'

    def __init__(
            self,
            compo_cls: 'rai.typing.CompoType',
            **kwargs: Any
            ) -> None:

        self.kwargs = kwargs
        self.compo_cls = compo_cls

    def __call__(self, **kwargs: Any) -> 'rai.typing.Compo':
        """Finish creating the partially created Compo."""
        kwargs2 = copy(self.kwargs)
        kwargs2.update(kwargs)
        return self.compo_cls(**kwargs2)




"""Type hints for RAIMAD.

These are used internally,
and you may also use them to annotate your RAIMAD packages.
"""

try:
    from typing import TypeAlias
except ImportError:
    # py3.9 and lower
    from typing_extensions import TypeAlias

import raimad as rai

CompoType: TypeAlias = type[rai.Compo]
Partial: TypeAlias = rai.Partial
CompoTypeLike: TypeAlias = type[rai.Compo] | rai.Partial

Compo: TypeAlias = rai.Compo
Proxy: TypeAlias = rai.Proxy
CompoLike: TypeAlias = rai.Compo | rai.Proxy

Point: TypeAlias = tuple[float, float]
BoundPoint: TypeAlias = rai.BoundPoint
PointLike: TypeAlias = Point | rai.BoundPoint

Poly: TypeAlias = list[Point]
Polys: TypeAlias = list[Poly]
Geoms: TypeAlias = dict[str, Polys]
Transform: TypeAlias = rai.Transform
BBox: TypeAlias = rai.BBox
BoundBBox: TypeAlias = rai.BoundBBox
Affine: TypeAlias = tuple[
    tuple[float, float, float],
    tuple[float, float, float],
    tuple[float, float, float],
    ]

LMapShorthand: TypeAlias = None | str | dict[str, str]
LMap: TypeAlias = rai.LMap




"""boundbbox.py: contains BBox class."""
import raimad as rai

class BoundBBox(rai.AbstractBBox['rai.typing.BoundPoint']):
    """
    BoundBBox: a BBox bound to a Proxy.

    A BoundBBox is just like a BBox, except the points
    returned by its attributes are BoundPoints,
    thus allowing you to perform transformations around
    bbox points of a proxy, for example like this:
    `someproxy.bbox.mid_right.to(somepoint)`
    """

    _proxy: 'rai.typing.Proxy'

    def __init__(
            self,
            proxy: 'rai.typing.Proxy',
            xyarray: 'rai.typing.Poly | None' = None,
            ) -> None:
        super().__init__(xyarray)
        self._proxy = proxy

    def interpolate(
            self,
            x_ratio: float,
            y_ratio: float,
            ) -> 'rai.typing.BoundPoint':
        """
        Find a point inside (or outside) the bbox given X and Y ratios.

        So, for example, 0,0 is top left,
        1,1 is bottom right, 0.5,0.5 is center,
        and 0.5,1 is bottom middle.
        The ratios may be negative or higher than 1,
        but doing so would probably make your code difficult to understand.

        Parameters
        ----------
        x_ratio: float
            A number, such that 0 represents all the way to the left
            of the bbox, and 1 represents all the way to the right.
        y_ratio: float
            A number, such that 0 represents all the way to the bottom
            of the bbox, and 1 represents all the way to the top.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        Returns
        -------
        rai.typing.BoundPoint
            The interpolated point.
        """
        point = super()._interpolate(x_ratio, y_ratio)
        return rai.BoundPoint(
            point[0],
            point[1],
            self._proxy
            )




import unittest

import raimad as rai
import cift

from .utils import GeomsEqual

class TestCIFNoReuse(GeomsEqual, unittest.TestCase):

    def test_cif_noreuse_rect(self):
        """
        """
        compo = rai.RectLW(10, 20)

        exporter = rai.cif.NoReuse(
            compo,
            multiplier=1,
            )

        parser = cift.Parser()
        parser.parse(exporter.cif_string)

        self.assertGeomsEqual(
            parser.layers,
            {
                'Lroot': [
                    [
                        (-5, -10),
                        (5, -10),
                        (5, 10),
                        (-5, 10),
                        ],
                    ]
                }
            )

    def test_cif_noreuse_proxy(self):
        """
        """
        compo = (
            rai.RectLW(10, 20)
            .proxy()
            .map('mylayer')
            .bbox.bot_left.to((0, 0))
            )

        exporter = rai.cif.NoReuse(
            compo,
            multiplier=1,
            )

        parser = cift.Parser()
        parser.parse(exporter.cif_string)

        self.assertGeomsEqual(
            parser.layers,
            {
                'Lmylayer': [
                    [
                        (0, 0),
                        (10, 0),
                        (10, 20),
                        (0, 20),
                        ],
                    ]
                }
            )


if __name__ == '__main__':
    unittest.main()




"""rectlw.py: home to RectLW compo."""

import raimad as rai

class RectLW(rai.Compo):
    """A rectangle defined by length and width."""

    browser_name = "RectLW"
    browser_tags = ["builtin", "polygon"]

    class Options:
        length = rai.Option('Length of rectangle', browser_default=15)
        width = rai.Option('Width of rectangle', browser_default=10)

    def _make(self, length: float, width: float) -> None:
        self.length = length
        self.width = width

        self.geoms.update({
            'root': [
                [
                    (- length / 2, - width / 2),
                    (+ length / 2, - width / 2),
                    (+ length / 2, + width / 2),
                    (- length / 2, + width / 2),
                    ]
                ]
            })




"""RAIMAD Command-line interface (CLI)."""

import argparse
from typing import Sequence
import sys
import os

import raimad as rai

ACTION_EXPORT = 'export'
ACTION_SHOW = 'show'
FILE_STDOUT = '-'

def cli(custom_args: Sequence[str] | None = None) -> None:
    """Parse command line arguments and execute the desired action."""

    ensure_pwd_in_path()

    parser = _setup_parser()
    if custom_args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(custom_args)

    if args.action == ACTION_EXPORT:
        _process_args_export(args)
        compo = args.component()

        rai.export_cif(compo, args.output_file)

    elif args.action == ACTION_SHOW:
        rai.show(args.component(), args.ignore_running)

    else:
        # This should never happen, since
        # argparse validates this.
        parser.error('Unknown action')

def _setup_parser() -> argparse.ArgumentParser:
    """Create the root parser for the CLI interface."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

    subparsers = parser.add_subparsers(
        title='Action',
        dest='action',
        description='Which action to perform',
        )

    _add_export_action(subparsers)
    _add_show_action(subparsers)

    return parser

def _add_export_action(
        # Can someone explain to me what this generic actually does
        # and why it crashes if it's not quoted?
        subparsers: 'argparse._SubParsersAction[argparse.ArgumentParser]'
        ) -> None:
    """Add the export action to the root parser."""
    parser = subparsers.add_parser(
        ACTION_EXPORT,
        help=(
            "Export a component to disk"
            )
        )

    parser.add_argument(
        'component',
        type=lambda string: rai.string_import(string, multiple=False),
        help=(
            'Component to export in the format '
            '`{dot-separated path to module}:{class name}`. '
            'For example, `raimad:Snowman`.'
            )
        )

    parser.add_argument(
        '--output-file',
        '-o',
        type=str,
        nargs='?',
        help=(
            f'Output file. Use `{FILE_STDOUT}` for stdout. '
            '`{name}` will get formatted with component name. '
            ),
        default='{name}.cif',
        )

def _add_show_action(
        subparsers: 'argparse._SubParsersAction[argparse.ArgumentParser]'
        ) -> None:
    """Add the show action to the root parser."""
    parser = subparsers.add_parser(
        ACTION_SHOW,
        help=(
            "Export a component and open it in a CIF viewer"
            )
        )

    parser.add_argument(
        'component',
        type=lambda string: rai.string_import(string, multiple=False),
        help=(
            'Component to export in the format '
            '`{dot-separated path to module}:{class name}`. '
            'For example, `raimad:Snowman`.'
            )
        )

    parser.add_argument(
        '--ignore-running',
        '-i',
        action='store_true',
        help="Launch the CIF viewer again even if it is already running",
        )


def _process_args_export(args: argparse.Namespace) -> None:
    args.output_file = args.output_file.replace(
        '{name}',
        args.component.__name__
        )

def ensure_pwd_in_path():
    pwd = os.getcwd()
    if pwd not in sys.path:
        sys.path.insert(0, pwd)




"""option.py: home to the Option class and its subclasses."""
from typing import ClassVar, Any

try:
    from typing import Self
except ImportError:
    # py3.10 and lower
    from typing_extensions import Self

import raimad as rai


class Option(rai.Annotation):
    """
    Option: an annotation for an option.

    Option objects are meant to be declared inside the `Options`
    nested class of every Compo.
    They serve as annotations for the options that the `_make`
    method takes that better explain what those options do.

    TODO explain that its better to use the subclasses
    """

    annot: str | rai.EmptyType | type
    default: Any | rai.EmptyType

    def __init__(
            self,
            desc: str,
            browser_default: Any | rai.EmptyType = rai.Empty) -> None:
        self.desc = desc
        self.category = type(self)
        self.name = 'this should be set in compo.__init_subclass_'
        self.default = 'this should be set in compo.__init_subclass_'
        self.browser_default = browser_default
        self.annot = 'this should be set in compo.__init_subclass_'

    # These are shorthands which will be assigned at
    # the end of this file.
    # The annotations are here to make mypy shut up
    Functional: ClassVar[type[Self]]
    Geometric: ClassVar[type[Self]]
    Debug: ClassVar[type[Self]]
    Environmental: ClassVar[type[Self]]

class Geometric(Option):
    """
    Geometric option annotation.

    This subclass of Option is meant for options that directly
    correspond to physical dimensions in the compo:
    lengths, widths, angles.
    """

class Functional(Option):
    """
    Functional option annotation.

    This subclass of Option is meant for options that correspond
    to functional properties of a compo:
    frequencies, impedances, etc.
    The idea is that Functional options are used to derive
    the geometry of a Compo
    """

class Debug(Option):
    """
    Debug option annotation.

    This subclass of Option is meant for options that
    are useful for debugging, but not production use.
    They can control whether extra info is printed to stderr,
    whether parts of the compo are ommitted for faster compilation,
    etc.
    """

class Environmental(Option):
    """
    Environmental option annotation.

    This subclass of Option is meant for options that
    prescribe envrionmental constants like temperature
    of absolute zero,
    planck length,
    and so forth,
    that you don't feel like hardcoding.
    """


Option.Functional = Functional
Option.Geometric = Geometric
Option.Debug = Debug
Option.Environmental = Environmental




"""
Test `rai.string_import` method,
a function that taskes a string in the format
{period-separated modules}.{class name}
and imports the corresponding RAIMAD component class
"""

import unittest
from pathlib import Path
import os
import tempfile
import sys

import raimad as rai

class TestStringImport(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.snowman_cif = rai.export_cif(rai.Snowman())

    def test_string_import_single(self):
        """
        Test importing a single component
        """
        self.assertIs(
            rai.Snowman,
            rai.string_import('raimad:Snowman', multiple=False)
            )

    def test_string_import_multiple(self):
        """
        Test importing all components from a module
        """
        # Dear future maintainer:
        # if you added a new built-in component to RAIMAD
        # and this test started failing,
        # just add your new component to the list below
        self.assertEqual(
            {
                rai.Snowman,
                rai.AnSec,
                rai.RectLW,
                rai.Circle,
                rai.RectWire,
                rai.CustomPoly,
                },
            set(
                rai.string_import('raimad', multiple=True)
                )
            )

    def test_string_import_single_fail_compo(self):
        """
        Test that the correct error is thrown importing a component that
        doesn't exist
        """
        with self.assertRaises(rai.err.StringImportError):
            rai.string_import('raimad:ComponentDoesNotExist')

    def test_string_import_single_fail_module(self):
        """
        Test that the correct error is thrown importing a component from
        a module that does not exist
        """
        with self.assertRaises(rai.err.StringImportError):
            rai.string_import('module_does_not_exist:Snowman')

    def test_string_import_single_fail_multiple(self):
        """
        Test that the correct error is thrown when you try
        to import multiple components with `multiple=False`
        """
        with self.assertRaises(rai.err.StringImportError):
            rai.string_import('raimad')

    def test_string_import_single_as_multiple(self):
        """
        Test that no error is thrown is `multiple=False`,
        and the string imports all components from a module,
        but the module has only one component.
        """
        with tempfile.TemporaryDirectory() as folder:
            os.chdir(folder)
            # This test will break pycoverage,
            # because pycoverage will think that `mymodule`
            # is a real module that needs evaluating,
            # but can't find its source.
            # Run pycoverage with `-i` flag.
            Path('mymodule.py').write_text(
                "import raimad as rai\n"
                "\n"
                "class MyCompo(rai.Compo):\n"
                "    def _make(self):\n"
                "        pass"
                )

            # Add the temporary directory to sys.path
            # so that the import machinery
            # can see `mymodule.py`
            sys.path.insert(0, folder)

            compo = rai.string_import("mymodule", multiple=False)

            # remove the temporary directory from sys.path
            # to avoid weirdness down the line.
            sys.path.pop(0)

            self.assertTrue(issubclass(compo, rai.Compo))


if __name__ == '__main__':
    unittest.main()




"""bbox.py: contains BBox class and relevant Exceptions."""

from typing import Iterator, Generic, TypeVar
from copy import copy

try:
    from typing import Self
except ImportError:
    # py3.10 and lower
    from typing_extensions import Self

import raimad as rai

# Disable to use computer-style Y axis (increases downwards)
# DO NOT TOUCH THIS IT WILL BREAK EVERYTHING
MATHEMATIC_Y_AXIS = True

class EmptyBBoxError(Exception):
    """
    EmptyBBoxError: attempted operation on an empty bbox.

    An empty bbox is one that does not contain any points.
    Most properties of an empty bbox (length, width, specific corners, etc.)
    do not make sense, and attempting to query them will raise this exception.
    """


T = TypeVar('T', 'rai.typing.Point', 'rai.typing.BoundPoint', )
class AbstractBBox(Generic[T]):
    """
    """

    max_x: float
    max_y: float
    min_x: float
    min_y: float

    def __init__(
            self,
            xyarray: 'rai.typing.Poly | None' = None,
            ):
        """
        Create a new BBox.

        Parameters
        ----------
        xyarray: np.ndarray | None
            A N x 2 numpy array containing points to initialize the bbox with.
            This is optional.
        proxy: rai.Proxy | None
            The proxy that this bbox should be bound to.
            This is optional.
        """
        self.max_x = float('-inf')
        self.max_y = float('-inf')
        self.min_x = float('inf')
        self.min_y = float('inf')

        if xyarray is not None:
            self.add_xyarray(xyarray)

    def as_list(self) -> list[float]:
        """Return as list of [min_x, min_y, max_x, max_y]."""
        return [self.min_x, self.min_y, self.max_x, self.max_y]

    def __iter__(self) -> Iterator[float]:
        """Return as iter of [min_x, min_y, max_x, max_y]."""
        return iter(self.as_list())

    def is_empty(self) -> bool:
        """
        Check if bbox is empty.

        An empoty bbox has no paints.
        BBoxes created with no `xyarray` argument are empty
        until you manually add points to them with
        `add_xyarray` or `add_point`.

        Returns
        -------
        bool
            Whether or not the bbox is empty.

        See Also
        --------
        .5        -------
        BBox.assert_nonempty()
        """
        return (
            self.max_x == float('-inf') or
            self.max_y == float('-inf') or
            self.min_x == float('inf') or
            self.min_y == float('inf')
            )

    def assert_nonempty(self, message: str | None = None) -> None:
        """
        Throw EmptyBBoxError is the bbox is empty, do nothing otherwise.

        Parameters
        ----------
        message : str
            Message to pass to EmptyBBoxError

        Raises
        ------
        EmptyBBoxError

        See Also
        --------
        BBox.is_empty()
        """
        if self.is_empty():
            if message is None:
                raise EmptyBBoxError()
            raise EmptyBBoxError()

    def copy(self) -> Self:
        """Copy bbox."""
        # TODO read up on how to properly do copies
        # in Python
        return copy(self)

    def add_xyarray(self, xyarray: 'rai.typing.Poly') -> None:
        """
        Add new points to the bounding box.

        Parameters
        ----------
        xyarray: np.ndarray
            A N x 2 numpy array containing points to add to the bbox.
        """
        for point in xyarray:
            self.add_point(point)

    def add_point(self, point: 'rai.typing.PointLike') -> None:
        """
        Add a new point to the bounding box.

        Parameters
        ----------
        point
            the point to add
        """
        # TODO simplify with ELIFs?
        if point[0] > self.max_x:
            self.max_x = point[0]
        if point[0] < self.min_x:
            self.min_x = point[0]

        if point[1] > self.max_y:
            self.max_y = point[1]
        if point[1] < self.min_y:
            self.min_y = point[1]

    @property
    def length(self) -> float:
        """
        Get length of the bbox.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        Returns
        -------
        float
            length of the bbox
        """
        self.assert_nonempty("Tried to get length of empty bbox.")
        return self.max_x - self.min_x

    @property
    def width(self) -> float:
        """
        Get width of the bbox.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        Returns
        -------
        float
            width of the bbox
        """
        self.assert_nonempty("Tried to get width of empty bbox.")
        return self.max_y - self.min_y

    @property
    def left(self) -> float:
        """
        Get the X coordinate of the left wall of the bbox.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        Returns
        -------
        float
            left X ccordinate of the bbox
        """
        self.assert_nonempty("Tried to get left of empty bbox.")
        return self.min_x

    @property
    def top(self) -> float:
        """
        Get the Y coordinate of the top wall of the bbox.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        Returns
        -------
        float
            top Y coordinate of the bbox
        """
        self.assert_nonempty("Tried to get top of empty bbox.")
        return self.max_y if MATHEMATIC_Y_AXIS else self.min_y

    @property
    def right(self) -> float:
        """
        Get the X coorinate of the right wall of the bbox.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        Returns
        -------
        float
            right X coordinate of the bbox
        """
        self.assert_nonempty("Tried to get right of empty bbox.")
        return self.max_x

    @property
    def bottom(self) -> float:
        """
        Get the Y coordinate of the bottom of the bbox.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        Returns
        -------
        float
            bottom Y of the bbox
        """
        self.assert_nonempty("Tried to get bottom of empty bbox.")
        return self.min_y if MATHEMATIC_Y_AXIS else self.max_y

    def _interpolate(
            self,
            x_ratio: float,
            y_ratio: float,
            ) -> 'rai.typing.Point':
        x = self.min_x + self.length * x_ratio
        y = self.min_y + self.width * y_ratio

        return (x, y)

    def interpolate(
            self,
            x_ratio: float,
            y_ratio: float,
            ) -> T:
        """
        Find a point inside (or outside) the bbox given X and Y ratios.

        So, for example, 0,0 is top left,
        1,1 is bottom right, 0.5,0.5 is center,
        and 0.5,1 is bottom middle.
        The ratios may be negative or higher than 1,
        but doing so would probably make your code difficult to understand.

        Parameters
        ----------
        x_ratio: float
            A number, such that 0 represents all the way to the left
            of the bbox, and 1 represents all the way to the right.
        y_ratio: float
            A number, such that 0 represents all the way to the bottom
            of the bbox, and 1 represents all the way to the top.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        """
        raise NotImplementedError()

    @property
    def mid(self) -> T:
        """
        Get the point in the MIDDLE of the bbox.

        Like this:
        .-.-.
        |   |
        . X .
        |   |
        ._._.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        """
        self.assert_nonempty("Tried to get middle point of empty bbox")
        return self.interpolate(0.5, 0.5)

    @property
    def top_mid(self) -> T:
        """
        Get the point in the TOP MIDDLE of the bbox.

        Like this:
        .-X-.
        |   |
        . . .
        |   |
        ._._.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        """
        self.assert_nonempty("Tried to get top middle point of empty bbox")
        return self.interpolate(0.5, MATHEMATIC_Y_AXIS)

    @property
    def bot_mid(self) -> T:
        """
        Get the point in the BOTTOM MIDDLE of the bbox.

        Like this:
        .-.-.
        |   |
        . . .
        |   |
        ._X_.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        """
        self.assert_nonempty("Tried to get bottom middle point of empty bbox")
        return self.interpolate(0.5, not MATHEMATIC_Y_AXIS)

    @property
    def mid_left(self) -> T:
        """
        Get the point in the MIDDLE LEFT of the bbox.

        Like this:
        .-.-.
        |   |
        X . .
        |   |
        ._._.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        """
        self.assert_nonempty("Tried to get middle left point of empty bbox")
        return self.interpolate(0, 0.5)

    @property
    def mid_right(self) -> T:
        """
        Get the point in the MIDDLE RIGHT of the bbox.

        Like this:
        .-.-.
        |   |
        . . X
        |   |
        ._._.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        """
        self.assert_nonempty("Tried to get middle right point of empty bbox")
        return self.interpolate(1, 0.5)

    @property
    def top_left(self) -> T:
        """
        Get the point in the TOP LEFT of the bbox.

        Like this:
        X-.-.
        |   |
        . . .
        |   |
        ._._.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        """
        self.assert_nonempty("Tried to get top left point of empty bbox")
        return self.interpolate(0, MATHEMATIC_Y_AXIS)

    @property
    def top_right(self) -> T:
        """
        Get the point in the TOP RIGHT of the bbox.

        Like this:
        .-.-X
        |   |
        . . .
        |   |
        ._._.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        """
        self.assert_nonempty("Tried to get top right point of empty bbox")
        return self.interpolate(1, MATHEMATIC_Y_AXIS)

    @property
    def bot_left(self) -> T:
        """
        Get the point in the BOTTOM LEFT of the bbox.

        Like this:
        .-.-.
        |   |
        . . .
        |   |
        X_._.

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        """
        self.assert_nonempty("Tried to get bottom left point of empty bbox")
        return self.interpolate(0, not MATHEMATIC_Y_AXIS)

    @property
    def bot_right(self) -> T:
        """
        Get the point in the middle of the bbox.

        Like this:
        .-.-.
        |   |
        . . .
        |   |
        ._._X

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        """
        self.assert_nonempty("Tried to get bottom right point of empty bbox")
        return self.interpolate(1, not MATHEMATIC_Y_AXIS)

    def pad(
            self,
            x: float = 0,
            y: float | None = None,
            /,
            *,
            left: float = 0,
            top: float = 0,
            right: float = 0,
            bottom: float = 0,
            ) -> Self:
        """
        Create a copy of this bbox with extra padding.

        Bound bboxes will be turned into unbound ones (Or will they O_O ).

        Examples
        --------
        bbox.pad(5)  # Pad evenly on every side
        bbox.pad(5, 10)  # Pad 5 units horizontally and 10 units vertically
        bbox.pad(left=10)  # Pad 10 units on the left side only
        bbox.pad(left=10, bottom=5)  # Pad 10 on th left, 5 on the bottom
        bbox.pad(5, left=10)  # Pad 15 units on the left, 5 everywhere else.

        Parameters
        ----------
        x: float
            Base padding (is y in not specified) or horizontal padding
            (if y IS specified)
        y: float
            Vertical padding.
        left: float
            Additional left padding
        right: float
            Additional right padding
        top: float
            Additional top padding
        bottom: float
            Additional bottom padding

        Raises
        ------
        EmptyBBoxError
            if the bbox is empty

        Returns
        -------
        Self
            A new bbox with the corresponding padding.
        """
        self.assert_nonempty("Tried to pad empty bbox")

        y = y or x

        new = self.copy()
        new.min_x -= left + x
        new.max_x += right + x
        new.min_y -= bottom + y
        new.max_y += top + y

        return new


class BBox(AbstractBBox['rai.typing.Point']):
    def interpolate(
            self,
            x_ratio: float,
            y_ratio: float,
            ) -> 'rai.typing.Point':
        return super()._interpolate(x_ratio, y_ratio)




"""circle.py: home to Circle builtin component."""

from math import pi
import raimad as rai

class Circle(rai.Compo):
    """A polygon that approximates a circle."""

    browser_tags = ["builtin", "polygon"]

    class Options:
        radius = rai.Option('Circle radius', browser_default=15)
        num_points = rai.Option('Number of points')

    def _make(self, radius: float, num_points: int = 200) -> None:

        self.geoms.update({
            'root': [
                [
                    rai.polar(angle / num_points * 2 * pi, radius)
                    for angle
                    in range(0, num_points)
                    ]
                ]
            })

        self.marks.center = (0, 0)




"""
Test the custom_base helper
"""

import unittest
import re

import raimad as rai

class TestCustomBase(unittest.TestCase):

    def test_custom_base(self):
        self.assertEqual(
            rai.custom_base(69420, ['0', '1']),
            bin(69420).lstrip('0b')
            )

        self.assertEqual(
            rai.custom_base(10, list('0123456789abcdef')),
            hex(10).lstrip('0x')
            )

        self.assertEqual(
            rai.custom_base(16, list('0123456789abcdef')),
            hex(16).lstrip('0x')
            )

        self.assertEqual(
            rai.custom_base(69420, list('0123456789abcdef')),
            hex(69420).lstrip('0x')
            )


if __name__ == '__main__':
    unittest.main()




import unittest

import raimad as rai

class TestSnowman(unittest.TestCase):

    def test_snowman(self):
        compo = rai.Snowman(nose_length=20, eye_size=2.5)
        geom = compo.steamroll()

        # Test layers
        self.assertEqual(geom.keys(), {'snow', 'carrot', 'pebble'})

        # Test three circles on snow layer
        self.assertEqual(len(geom['snow']), 3)

        # Test marks
        self.assertTrue('nose' in compo.marks.keys())


if __name__ == '__main__':
    unittest.main()




"""proxy.py: home to Proxy class and supporting classes."""

from typing import Iterator

try:
    from typing import Self
except ImportError:
    # py3.10 and lower
    from typing_extensions import Self

from copy import copy

import raimad as rai

class LMap:
    """
    LMap: Layer map.

    LMaps tell RAIMAD how the layers of a subcompo map
    to its parent compo.
    An LMap is canonically represented by an LMapShorthand,
    which may be either None, a string, or a dict.

    A None layermap is "transparent":
    the subcompo layers are mapped directly to the parent.
    If the subcompo has layers "foo" and "bar",
    then the parent will have layers "foo" and "bar".

    A string layermap collapses all of the subcompo
    layers to one layer in the parent.
    If the subcompo has layers "foo" and "bar",
    and a layermap of "baz",
    then the parent will have a layer "baz" containing
    the geometries from both "foo" and "bar".

    A dict layermap allows arbitrary layer mapping.
    The keys of the dict correspond to subcompo layers,
    and the values correspond to parent layers

    TODO explain edgecases.
    """

    def __init__(self, shorthand: 'rai.typing.LMapShorthand') -> None:
        self.shorthand = shorthand

    def __getitem__(self, name: str) -> str:
        """Given a child layer, return the parent layer."""
        if self.shorthand is None:
            return name

        if isinstance(self.shorthand, str):
            return self.shorthand

        if isinstance(self.shorthand, dict):
            return self.shorthand[name]

        assert False
    # TODO hacky

    def copy(self) -> Self:
        """Copy this lmap."""
        return type(self)(self.shorthand)

    def compose(self, other: Self) -> Self:
        """
        Merge this lmap with another lmap.

        Parameters
        ----------
        other: Self
            The other lmap

        Returns
        -------
        Self
            Self is returned to allow method chaining.
        """
        if other.shorthand is None:
            # None lmap, pass everything through
            pass

        elif isinstance(other.shorthand, str):
            # single-string lmap, flatten everything to one layer
            self.shorthand = other.shorthand

        elif isinstance(other.shorthand, dict):
            # TODO better way to do this?
            if isinstance(self.shorthand, dict):
                for self_k, self_val in tuple(self.shorthand.items()):
                    try:
                        self.shorthand[self_k] = other[self_val]
                    except KeyError:
                        pass

            elif self.shorthand is None:
                self.shorthand = copy(other.shorthand)

            elif isinstance(self.shorthand, str):
                self.shorthand = other[self.shorthand]

        else:
            assert False

        return self

class ProxiedMarksContainer(
        rai.FilteredDictList[
            'rai.typing.Point',
            'rai.typing.PointLike',
            'rai.typing.BoundPoint',
            ]):
    """
    ProxiedMarksContainer: A MarksContainer seen through a Proxy.

    Unlike a MarksContainer, which returns regular Points,
    this ProxiedMarksContainer returns BoundPoints that
    are bound to a proxy.
    This allows using marks as a reference point for
    transforms,
    for example
    `someproxy.marks.left_corner.rotate(pi)`
    to rotate around the left corner.
    """

    def __init__(
            self,
            proxy: 'rai.typing.Proxy',
            dict_: dict[int | str, 'rai.typing.Point'] | None = None,
            *,
            copy: bool | None = None
            ):
        super().__init__(dict_, copy=copy)
        self._proxy = proxy

    def _filter_set(self, val: 'rai.typing.PointLike') -> 'rai.typing.Point':
        """
        Set filter for markscontainer.

        Make sure that no matter what creature gets passed in
        (regular tuple or boundpoint or whatever),
        what gets stored is a simple regular tuple.
        """
        return (val[0], val[1])

    def _filter_get(self, val: 'rai.typing.Point') -> 'rai.typing.BoundPoint':

        assert not isinstance(val, rai.BoundPoint)
        assert isinstance(val, tuple)
        assert type(val) is tuple
        assert len(val) == 2

        point = self._proxy.transform_point(val)
        boundpoint = rai.BoundPoint(
            point[0],
            point[1],
            self._proxy
            )
        return boundpoint


class Proxy:
    """
    Proxy: transforming and layermapping wrapper around a Compo.

    TODO description here.
    """

    compo: 'rai.typing.CompoLike'

    def __init__(self,
                 compo: 'rai.typing.CompoLike',
                 lmap: 'rai.typing.LMapShorthand' = None,
                 transform: 'rai.typing.Transform | None' = None,
                 _cif_link: bool = False,
                 _autogen: bool = False,
                 ):
        self._cif_linked = False
        self._cif_link = _cif_link
        self._autogen = _autogen
        self.compo = compo
        self.lmap = LMap(lmap)
        self.transform = transform or rai.Transform()

    def steamroll(self) -> 'rai.typing.Geoms':
        """
        Get all geometries of this proxy.

        This function returns all geometries
        (i.e. all raw geometries as well as subcompos)
        of the CompoLike pointed to by this proxy,
        as seen through this proxy.
        """
        return {
            self.lmap[layer]: [
                self.transform.transform_xyarray(geom)
                for geom in geoms
                ]
            for layer, geoms
            in self.compo.steamroll().items()
            }

    def get_flat_transform(self, maxdepth: int = -1) -> 'rai.typing.Transform':
        """Get flattened transform from tower of proxies."""
        if maxdepth == 0 or isinstance(self.compo, rai.Compo):
            return self.transform.copy()

        # return (
        #     self.compo.get_flat_transform(maxdepth - 1)
        #     .compose(self.transform)
        #     )

        # TODO huh? copy?
        return self.transform.copy().compose(
            self.compo.get_flat_transform(maxdepth - 1)
            )

    def get_flat_lmap(self, maxdepth: int = -1) -> LMap:
        """Get flattened lmap from tower of proxies."""
        if maxdepth == 0 or isinstance(self.compo, rai.Compo):
            return self.lmap.copy()

        return self.lmap.copy().compose(
            self.compo.get_flat_lmap(maxdepth - 1)
            )

    @property
    def geoms(self) -> 'rai.typing.Geoms':
        """
        Get the raw geometries as seen through this proxy.

        This function returns the raw geometries
        (i.e. NOT geometries defined in subcompos)
        defined in the CompoLike pointed to by this Proxy,
        as seen through this proxy.
        """
        return {
            self.lmap[layer]: [
                self.transform.transform_xyarray(geom)
                for geom in geoms
                ]
            for layer, geoms
            in self.compo.geoms.items()
            }

    @property
    def subcompos(self) -> rai.SubcompoContainer:
        """
        Get subcompos of the CompoLike pointed to by this proxy.

        A special `SubcompoContainer` is returned
        that is aware that you're looking at it through a proxy,
        so all of the subcompos
        will be wrapped in copies of this proxy.
        """
        return self.compo.subcompos._get_proxy_view(self)

    def final(self) -> 'rai.typing.Compo':
        """Return the compo at the bottom of a tower of proxies."""
        return self.compo.final()

    def depth(self) -> int:
        """
        Measure depth of a tower of proxies.

        The depth of a compo is 0,
        a proxy pointing to a compo is 1,
        a proxy pointing to a proxy pointing to a compo is 2,
        and so forth.
        """
        return self.compo.depth() + 1

    def descend(self) -> 'Iterator[rai.typing.CompoLike]':
        """Descend a tower of proxies to the compo at the bottom."""
        yield self
        yield from self.compo.descend()

    def descend_p(self) -> 'Iterator[rai.typing.Proxy]':
        """Descend a tower of proxies to the lowest proxy."""
        yield self
        yield from self.compo.descend_p()

    def proxy(self) -> 'rai.typing.Proxy':
        """Return a new proxy pointing to this proxy."""
        return rai.Proxy(self)

    def walk_hier(self) -> 'Iterator[rai.typing.Proxy]':
        """
        Traverse the subcomponent hierarchy of the CompoLike of this proxy.

        This method will recursively walk through the entire
        subcompo hierarchy of the CompoLike pointed to by
        this proxy.
        Each node is wrapped in a copy of this proxy.
        """
        # TODO test this?
        for subcompo in self.compo.walk_hier():
            yield self.copy_reassign(subcompo, _autogen=True)

    def copy(self) -> 'rai.typing.Proxy':
        """
        Copy this proxy.

        The new proxy will have a copy of this proxies
        layermap and transform,
        but will point to the same CompoLike.

        Returns
        -------
        Self
            The new copied proxy

        Raises
        ------
        Exception
            We do not yet allow copying a proxy that points
            to a proxy,
            because it is confusing and we haven't
            thought of a good usecase.
            So if you try to `.copy()` a proxy that points
            to a proxy,
            there will be an error.
            This may change in the future.
        """
        if self.depth() > 1:
            # TODO think about why someone would want to do this
            raise Exception("Copying proxy of depth more than 1")

        return type(self)(
            self.compo,
            self.lmap.shorthand,
            self.transform.copy(),
            )

    def copy_reassign(
            self,
            new_subcompo: 'rai.typing.CompoLike',
            _autogen: bool = False,
            ) -> 'rai.typing.Proxy':
        """
        Copy this proxy and point it a different CompoLike.

        This method creates a new proxy with a copy
        of this proxy's transform and layermap,
        and points it to a new CompoLike.
        Note that in case of a tower of proxies
        (a proxy pointing to a proxy pointing to a proxy...),
        only the topmost proxy is copied.

        Parameters
        ----------
        new_subcompo: rai.typing.CompoLike
            Proxy or Compo that the copied proxy should point to

        _autogen: bool
            This is used to mark the copied proxy as "autogenerated",
            for example if `copy_reassign` was used by
            SubcompoContainer to create a new proxy.
            This has no effect on how a proxy functions,
            but can be useful for debugging.
            This parameter is intended to be set only
            by internal RAIMAD functions,
            not by the end user

        Returns
        -------
        Self
            The new proxy is returned.
        """
        return type(self)(
            new_subcompo,
            self.lmap.shorthand,
            self.transform.copy(),
            _autogen=_autogen,
            )

    def transform_point(
            self,
            point: 'rai.typing.PointLike'
            ) -> 'rai.typing.PointLike':
        """
        Apply this proxies transform to a point, return the transformed point.

        A Point (tuple of two floats) is always returned,
        even if a BoundPoint is passed in.

        Parameters
        ----------
        point : rai.typing.PointLike
            The point to transform.

        Returns
        -------
        rai.typing.Point
            The transformed point.
        """
        return self.transform.transform_point(
            self.compo.transform_point(point)
            )

    def __str__(self) -> str:
        """Return string representation of this proxy."""
        stack = ''.join([
            'ma'[proxy._autogen]
            for proxy in self.descend_p()
            ])
        return (
            "<"
            f"Proxy of {self.final()} at {rai.wingdingify(id(self))} "
            f"stack `{stack}x`"
            ">"
            )

    def __repr__(self) -> str:
        """Return string representation of this proxy."""
        return self.__str__()

    # Transform functions #
    # TODO for all transforms
    # TODO same implementation as in Compo
    # TODO stack another transform or modify self?
    def scale(self, factor: float) -> Self:
        """
        Scale this proxy.

        Parameters
        ----------
        x : float
            Factor to scale by along the x axis
        y : float
            Factor to scale by along the y axis.
            If unspecified or None,
            use the x scale factor.

        Returns
        -------
        Self
            self is returned to allow chaining methods.
        """
        self.transform.scale(factor)
        return self

    def movex(self, factor: float) -> Self:
        """
        Move this proxy horizontally.

        Parameters
        ----------
        x : float
            Move this many units along x axis.

        Returns
        -------
        Self
            self is returned to allow chaining methods.
        """
        self.transform.movex(factor)
        return self

    def movey(self, factor: float) -> Self:
        """
        Move this proxy vertically.

        Parameters
        ----------
        y : float
            Move this many units along y axis.

        Returns
        -------
        Self
            self is returned to allow chaining methods.
        """
        self.transform.movey(factor)
        return self

    def move(self, x: float, y: float) -> Self:
        """
        Move this proxy.

        Parameters
        ----------
        x : float
            Move this many units along x axis.
        y : float
            Move this many units along y axis.

        Returns
        -------
        Self
            self is returned to allow chaining methods.
        """
        self.transform.move(x, y)
        return self

    def hflip(self, x: float = 0) -> Self:
        """
        Flip (mirror) along horizontal axis.

        Parameters
        ----------
        x : float
            Flip around this horizontal line (x coordinate)

        Returns
        -------
        Self
            self is returned to allow chaining methods.
        """
        self.transform.hflip(x)
        return self

    def vflip(self, y: float = 0) -> Self:
        """
        Flip (mirror) along vertical axis.

        Parameters
        ----------
        y : float
            Flip around this vertical line (y coordinate)

        Returns
        -------
        Self
            self is returned to allow chaining methods.
        """
        self.transform.vflip(y)
        return self

    def flip(self, x: float = 0, y: float = 0) -> Self:
        """
        Flip (mirror) along both horizontal and vertical axis.

        Parameters
        ----------
        x : float
            Flip around this point (x coordinate). Default: origin
        y : float
            Flip around this point (y coordinate). Default: origin

        Returns
        -------
        Self
            self is returned to allow chaining methods.
        """
        self.transform.flip(x, y)
        return self

    def rotate(self, angle: float, x: float = 0, y: float = 0) -> Self:
        """
        Rotate this proxy, possibly around a point.

        Arguments
        ---------
        angle : float
            Rotate by this many radians in the positive orientation
        x : float
            Rotate around this point (x coord). Default: origin
        y : float
            Rotate around this point (y coord). Default: origin

        Returns
        -------
        Self
            self is returned to allow method chaining.
        """
        self.transform.rotate(angle, x, y)
        return self

    def map(self, lmap_shorthand: 'rai.typing.LMapShorthand') -> Self:
        """Apply layermap to proxy."""
        self.lmap = LMap(lmap_shorthand)
        return self

    @property
    def marks(self) -> ProxiedMarksContainer:
        """Get marks of proxy."""
        return ProxiedMarksContainer(
            self,
            self.compo.marks._dict,
            copy=False
            )

    # bbox functions #
    # TODO same as compo -- some sort of reuse?
    @property
    def bbox(self) -> 'rai.BoundBBox':
        """Get bbox of proxy."""
        bbox = rai.BoundBBox(proxy=self)
        for geoms in self.steamroll().values():
            for geom in geoms:
                bbox.add_xyarray(geom)
        return bbox

    # snapping functions #
    def snap_left(self, other: Self) -> Self:
        """
        Move this proxy so its bbox is to the left of the target proxy.

         ________
        |        |
        |        |________
        |        |        |
        |  self  | target |
        |        |________|
        |        |
        |________|

        Parameters
        ----------
        The target proxy

        Returns
        -------
        Self
            This proxy is returned to allow method chaining.
        """
        self.bbox.mid_right.to(other.bbox.mid_left)
        return self

    def snap_right(self, other: Self) -> Self:
        """
        Move this proxy so its bbox is to the right of the target proxy.

                  ________
                 |        |
         ________|        |
        |        |        |
        | target |  self  |
        |________|        |
                 |        |
                 |________|

        Parameters
        ----------
        The target proxy

        Returns
        -------
        Self
            This proxy is returned to allow method chaining.
        """
        self.bbox.mid_left.to(other.bbox.mid_right)
        return self

    def snap_above(self, other: Self) -> Self:
        """
        Move this proxy so its bbox is directly above the target proxy.

         ________________ 
        |                |
        |      self      |
        |________________|
            |        |
            | target |
            |________|

        Parameters
        ----------
        The target proxy

        Returns
        -------
        Self
            This proxy is returned to allow method chaining.
        """
        self.bbox.bot_mid.to(other.bbox.top_mid)
        return self

    # TODO other should be CompoLike not Proxy, Right??
    def snap_below(self, other: Self) -> Self:
        """
        Move this proxy so its bbox is directly below the target proxy.

             ________ 
            |        |
            | target |
         ___|________|___ 
        |                |
        |      self      |
        |________________|

        Parameters
        ----------
        The target proxy

        Returns
        -------
        Self
            This proxy is returned to allow method chaining.
        """
        self.bbox.top_mid.to(other.bbox.bot_mid)
        return self

    def _repr_svg_(self) -> str:
        """
        Make svg representation of component.

        This is not an official magic method specified by Python,
        but rather a convention used by Jupyter Notebook
        and related tools.
        We also use it in RAIMARK.
        """
        return rai.export_svg(self)




"""
Test for lmap
"""

import unittest

import raimad as rai

class TestLMap(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_lmap_compose_none_x_none(self):
        """
        """
        below = rai.LMap(None)
        above = rai.LMap(None)
        below.compose(above)

        self.assertEqual(below.shorthand, None)
        self.assertEqual(below['somelayer'], 'somelayer')
        self.assertEqual(below['someotherlayer'], 'someotherlayer')

    def test_lmap_compose_none_x_str(self):
        """
        """
        below = rai.LMap(None)
        above = rai.LMap('theonelayer')
        below.compose(above)

        self.assertEqual(below.shorthand, 'theonelayer')
        self.assertEqual(below['test1'], 'theonelayer')
        self.assertEqual(below['test2'], 'theonelayer')

    def test_lmap_compose_none_x_dict(self):
        """
        """
        below = rai.LMap(None)
        above = rai.LMap({
            'ayy': 'lmao',
            'foo': 'bar',
            })
        below.compose(above)

        self.assertEqual(below['ayy'], 'lmao')
        self.assertEqual(below['foo'], 'bar')
        with self.assertRaises(KeyError):
            # TODO custom error for lmap that derives from
            # keyerror?
            below['notpresent']

    def test_lmap_compose_str_x_none(self):
        """
        """
        below = rai.LMap('theonelayer')
        above = rai.LMap(None)
        below.compose(above)

        self.assertEqual(below.shorthand, 'theonelayer')
        self.assertEqual(below['first'], 'theonelayer')
        self.assertEqual(below['second'], 'theonelayer')

    def test_lmap_compose_str_x_str(self):
        """
        """
        below = rai.LMap('first')
        above = rai.LMap('theonelayer')
        below.compose(above)

        self.assertEqual(below.shorthand, 'theonelayer')
        self.assertEqual(below['first'], 'theonelayer')
        self.assertEqual(below['second'], 'theonelayer')

    def test_lmap_compose_str_x_dict(self):
        """
        """
        below = rai.LMap('theone')
        above = rai.LMap({
            'theone': 'theother',
            'notpresent': 'cantgethere',
            })
        below.compose(above)

        self.assertEqual(below['theone'], 'theother')
        # TODO unsure what the proper behavior should be
        # when getting `notpresent` or some completely other key

    def test_lmap_compose_dict_x_none(self):
        """
        """
        below = rai.LMap({
            'ayy': 'lmao',
            'foo': 'bar',
            })
        above = rai.LMap(None)
        below.compose(above)

        self.assertEqual(below['ayy'], 'lmao')
        self.assertEqual(below['foo'], 'bar')
        with self.assertRaises(KeyError):
            below['notpresent']
        with self.assertRaises(KeyError):
            below['bar']

    def test_lmap_compose_dict_x_str(self):
        """
        """
        below = rai.LMap({
            'ayy': 'lmao',
            'foo': 'bar',
            })
        above = rai.LMap('theone')
        below.compose(above)

        self.assertEqual(below['ayy'], 'theone')
        self.assertEqual(below['foo'], 'theone')
        self.assertEqual(below['rando'], 'theone')
        # TODO do we want to be able to specify a
        # "lenient dict" layermap that acts like a dict layermap,
        # but instead throwing error on unknown layer,
        # it just passes it thru?

    def test_lmap_compose_dict_x_dict(self):
        """
        """
        below = rai.LMap({
            'ayy': 'lmao',
            'foo': 'bar',
            'syn': 'ack',
            'yin': 'yang',
            })
        above = rai.LMap({
            'bar': 'baz',
            'ack': 'nack',
            })
        below.compose(above)

        self.assertEqual(below['ayy'], 'lmao')
        self.assertEqual(below['foo'], 'baz')
        self.assertEqual(below['syn'], 'nack')
        self.assertEqual(below['yin'], 'yang')


if __name__ == '__main__':
    unittest.main()




"""dictlist.py: home to FilteredDictList and DictList classes."""

from typing import (
    KeysView,
    Iterable,
    Iterator,
    TypeVar,
    Generic,
    Any,
    NoReturn,
    )

T_STORED = TypeVar('T_STORED')
T_ADDED = TypeVar('T_ADDED')
T_RETURNED = TypeVar('T_RETURNED')
class FilteredDictList(Generic[T_STORED, T_ADDED, T_RETURNED]):
    """
    Base class for making containers that act like both list and dict.

    FilteredDictList stores data in an internal dict,
    but that data can also be accessed by index,
    and by attribute notation.
    FilteredDictList is an abstract class with two virtual
    methods: _filter_set and _filter_get.
    _filter_set takes any new item added to the list
    (via assignment to attribute, assignment to key, or .append)
    of generic type T_ADDED and transforms it to
    generic type T_STORED,
    which is what is actually stored in the encapsulated dict.
    _filter_get takes an item of type T_STORED
    and transforms it into T_RETURNED just before it
    is returned from the list
    (via getting an attribute, key, by index, or by iteration)

    TODO raises, examples, link to dictlist
    TODO explain how it is used in markscontainer, etc.
    """

    _dict: dict[str | int, T_STORED]

    def __init__(
            self,
            dict_: dict[int | str, T_STORED] | None = None,
            *,
            copy: bool | None = None
            ) -> None:

        if dict_ is None:
            self._dict = dict()

        elif isinstance(dict_, dict):
            if copy is True:
                self._dict = dict(dict_)

            elif copy is False:
                self._dict = dict_

            elif copy is None:
                raise TypeError(
                    "Must specify `copy` (bool) if passing `dict_`"
                    )

            else:
                raise TypeError('`copy` must be a bool')

        else:
            raise TypeError('`dict_` must be a dict')

        self._post_init()

    def __setitem__(self, key: int | str, val: T_ADDED) -> None:
        """Add or replace item in dictlist by key."""
        if isinstance(key, int):
            self._dict[tuple(self._dict.keys())[key]] = self._filter_set(val)

        elif isinstance(key, str):
            if key.startswith('_'):
                raise KeyError(
                    "Keys cannot start with underscore; "
                    "underscores are reserved for custom attributes. "
                    )

            if hasattr(self.__class__, key):
                raise KeyError(
                    "Key cannot be the same as "
                    "an existing class attribute. "
                    )

            self._dict.__setitem__(key, self._filter_set(val))

        else:
            raise TypeError("key must be int or str")

        self._sanitycheck()

    def __getitem__(self, key: int | str) -> T_RETURNED:
        """Get item from dictlist by key."""
        if isinstance(key, int):
            val = list(self._dict.values())[key]
        elif isinstance(key, str):
            val = self._dict.__getitem__(key)
        else:
            raise KeyError("Key must be int or str")

        return self._filter_get(val)

    def __setattr__(self, name: str, val: T_ADDED) -> None:
        """Add item to dictlist by key with attr notation."""
        if name.startswith('_'):
            super().__setattr__(name, val)

        elif hasattr(self.__class__, name):
            self._sanitycheck()
            raise AttributeError(
                f"Cannot override class attribute {name}"
                )

        else:
            self._dict[name] = self._filter_set(val)

        self._sanitycheck()

    def __getattr__(self, name: str) -> T_RETURNED | Any:
        """Get item from dictlist by key with attr notation."""
        if name.startswith('_'):
            return super().__getattribute__(name)

        return self._filter_get(self._dict[name])

    def append(self, val: T_ADDED) -> None:
        """Add item to dictlist."""
        self._sanitycheck()
        self._dict[len(self)] = self._filter_set(val)

    def extend(self, items: Iterable[T_ADDED]) -> None:
        """Add multiple items to dictlist."""
        for item in items:
            self.append(item)

    def _sanitycheck(self) -> None:
        assert all((isinstance(key, (int, str)) for key in self._dict.keys()))

        int_keys = [key for key in self._dict.keys() if isinstance(key, int)]
        if int_keys:
            assert max(int_keys) <= len(self)

    # TODO update method

    def __iter__(self) -> NoReturn:
        """
        Deliberately not implemented.

        Iterating through a dictlist is ambiguous.
        Use `.keys()`, `.values()`, or `.items()`.
        """
        raise NotImplementedError(
            "Iterating over dictlist is ambiguous. "
            "Please use `.keys()`, `.values()`, or `.items()`."
            )

    def items(self) -> Iterator[tuple[str | int, T_RETURNED]]:
        """Get iterator of key-value pairs in DictList."""
        return (
            (key, self._filter_get(val))
            for key, val in self._dict.items()
            )

    def values(self) -> Iterator[T_RETURNED]:
        """Get values in DictList."""
        return (
            self._filter_get(val)
            for val in self._dict.values()
            )

    def keys(self) -> KeysView[str | int]:
        """Get keys of DictList."""
        return self._dict.keys()

    def __len__(self) -> int:
        """Get number of items stored in DictList."""
        return self._dict.__len__()

    def _post_init(self) -> None:
        """
        Function that gets automatically run after __init__.

        Child classes may override this method to add more code
        after running __init__ without having to remember
        __init__'s signature.
        """

    def _filter_set(self, val: T_ADDED) -> T_STORED:
        raise NotImplementedError

    def _filter_get(self, val: T_STORED) -> T_RETURNED:
        raise NotImplementedError


T = TypeVar('T')
class DictList(FilteredDictList[T, T, T]):
    """
    A container that acts like a dict, a list, and an object all in one.

    This is a simplified version of FilteredDictList
    where T_STORED, T_ADDED, and T_RETURNED are all the same.
    It is not an abstract class, you can instantiate it
    directly.

    TODO examples.
    """

    def _filter_set(self, val: T) -> T:
        return val

    def _filter_get(self, val: T) -> T:
        return val




import unittest
from contextlib import redirect_stderr
import shlex
import os
import subprocess
from io import StringIO
import tempfile
from pathlib import Path

import raimad as rai

class TestCLI(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.snowman_cif = rai.export_cif(rai.Snowman())

    def test_cli_export_cif_noargs(self):
        with tempfile.TemporaryDirectory() as folder:
            os.chdir(folder)
            subprocess.run(shlex.split(
                'python -m raimad export raimad:Snowman'
                ), check=True)

            cif_string = Path('Snowman.cif').read_text()

        self.assertEqual(self.snowman_cif, cif_string)

    def test_cli_export_cif_file(self):
        with tempfile.TemporaryDirectory() as folder:
            os.chdir(folder)
            subprocess.run(shlex.split(
                'python -m raimad export raimad:Snowman -o compo.cif'
                ), check=True)

            cif_string = Path('compo.cif').read_text()

        self.assertEqual(self.snowman_cif, cif_string)

# TODO other formats? svg?

if __name__ == '__main__':
    unittest.main()




"""docparse.py: utilities for parsing docstrings"""

from typing import NamedTuple
import inspect

class SplitDocstring(NamedTuple):
    heading: str
    description: str


def split_docstring(docstring: str | None) -> SplitDocstring:
    """
    Split docstring by first line and other lines
    """
    if not docstring:
        return SplitDocstring('', '')

    split = inspect.cleandoc(docstring).split('\n', 1)

    if len(split) == 0:
        return SplitDocstring('', '')

    elif len(split) == 1:
        return SplitDocstring(split[0], '')

    else:
        return SplitDocstring(split[0], split[1])




import unittest

import raimad as rai

class BareGeometric(rai.Compo):
    def _make(self):
        self.geoms.update({
            'root': [
                [
                    [0, 0],
                    [0, 10],
                    [10, 10],
                    [10, 0],
                    ],
                [
                    [20, 20],
                    [40, 20],
                    [30, 40],
                    ],
                ]
            })

class BareStructural(rai.Compo):
    def _make(self):
        self.subcompos.append(
            rai.Proxy(
                BareGeometric(),
                {'root': 'upper'},
                )
            )
        self.subcompos.append(
            rai.Proxy(
                BareGeometric(),
                {'root': 'lower'},
                )
            )

class TestCompo(unittest.TestCase):

    def test_bare_geometric(self):
        compo = BareGeometric()
        geom = compo.steamroll()

        # Test one root layer
        self.assertEqual(geom.keys(), {'root'})

        # Test two polys on the layer
        self.assertEqual(len(geom['root']), 2)

    def test_bare_geometric_copied(self):
        compo = BareGeometric()
        self.assertIsNot(compo.geoms, compo.steamroll())

    def test_bare_structural(self):
        compo = BareStructural()
        geom = compo.steamroll()

        # Test one root layer
        self.assertEqual(geom.keys(), {'upper', 'lower'})

        # Test two polys on the top lyaer
        self.assertEqual(len(geom['upper']), 2)

        # Test two polys on the top lyaer
        self.assertEqual(len(geom['lower']), 2)

    def test_add_invalid_subcompo(self):
        """
        Test that adding anything other than a Proxy
        as a subcomponent raises an error
        """
        class MyCompo(rai.Compo):
            def _make(self):
                self.subcompos.invalid = 'invalid'

        with self.assertRaises(rai.err.InvalidSubcompoError):
            MyCompo()

    def test_add_invalid_compo_as_subcompo(self):
        """
        Test that adding a Compo as a subcompo
        (instead of a Proxy) raises an error
        """
        class MyCompo(rai.Compo):
            def _make(self):
                self.subcompos.invalid = rai.Snowman()

        with self.assertRaises(rai.err.CompoInsteadOfProxyAsSubcompoError):
            MyCompo()

        # This tests the inheritance relationship between the two error types
        with self.assertRaises(rai.err.InvalidSubcompoError):
            MyCompo()

    def test_fail_transform_compo(self):
        """
        Test that trying to use transformation methods on a compo
        (move, scale, etc)
        raises an error telling you to transform a Proxy of it instead
        """
        compo = rai.Snowman()

        with self.assertRaises(rai.err.TransformCompoError):
            compo.move(10, 10)

        with self.assertRaises(rai.err.TransformCompoError):
            compo.movex(10)

        with self.assertRaises(rai.err.TransformCompoError):
            compo.movey(10)

        with self.assertRaises(rai.err.TransformCompoError):
            compo.scale(10, 10)

        with self.assertRaises(rai.err.TransformCompoError):
            compo.rotate(10)

        with self.assertRaises(rai.err.TransformCompoError):
            compo.hflip()

        with self.assertRaises(rai.err.TransformCompoError):
            compo.vflip()

        with self.assertRaises(rai.err.TransformCompoError):
            compo.flip()

    def test_fail_copy_compo(self):
        """
        Test that trying to copy a compo
        raises an error telling you to copy a Proxy of it instead
        """
        compo = rai.Snowman()

        with self.assertRaises(rai.err.CopyCompoError):
            compo.copy()

        # Test the inheritance relationship between the error types
        with self.assertRaises(rai.err.ProxyCompoConfusionError):
            compo.copy()


if __name__ == '__main__':
    unittest.main()




import unittest

import raimad as rai

from .utils import ArrayAlmostEqual

class Compo_direct(rai.Compo):
    def _make(self):
        self.subcompos.beam = rai.RectLW(2, 20).proxy()
        self.subcompos.coup_top = rai.RectLW(10, 2).proxy()
        self.subcompos.coup_bot = rai.RectLW(8, 2).proxy()

        self.subcompos.beam.bbox.mid.to((0, 0))

        self.subcompos.coup_top.snap_above(self.subcompos.beam)
        self.subcompos.coup_bot.snap_below(self.subcompos.beam)

class Compo_auto(rai.Compo):
    def _make(self):
        beam = rai.RectLW(2, 20).proxy().map('root')
        coup_top = rai.RectLW(10, 2).proxy().map('root')
        coup_bot = rai.RectLW(8, 2).proxy().map('root')

        beam.bbox.mid.to((0, 0))

        coup_top.snap_above(beam)
        coup_bot.snap_below(beam)

        self.auto_subcompos()

#class Compo_shorthand(rai.Compo):
#    def _make(self):
#        beam = self.subcompo(rai.RectLW(2, 20), 'beam')
#        coup_top = self.subcompo(rai.RectLW(10, 2), 'coup_top')
#        coup_bot = self.subcompo(rai.RectLW(8, 2), 'coup_bot')
#
#        beam.bbox.mid.to((0, 0))
#
#        coup_top.snap_above(beam)
#        coup_bot.snap_below(beam)


class TestNamedSubcompos(ArrayAlmostEqual, unittest.TestCase):

    def test_named_subcompos(self):

        compo = Compo_direct()

        self.assertArrayAlmostEqual(
            compo.subcompos.coup_top.bbox.top_mid,
            (0, 20 / 2 + 2)
            )

        self.assertEqual(
            len(compo.subcompos),
            3
            )

    def test_named_subcompos_auto(self):

        compo = Compo_auto()

        self.assertArrayAlmostEqual(
            compo.subcompos.coup_top.bbox.top_mid,
            (0, 20 / 2 + 2)
            )

        self.assertEqual(
            len(compo.subcompos),
            3
            )

    #def test_named_subcompos_shorthand(self):

    #    compo = Compo_shorthand()

    #    self.assertArrayAlmostEqual(
    #        compo.subcompos.coup_top.bbox.top_mid,
    #        (0, 20 / 2 + 2)
    #        )

    #    self.assertEqual(
    #        len(compo.subcompos),
    #        3
    #        )


if __name__ == '__main__':
    unittest.main()




"""rectwire.py: home to RectWire compo and relevant exceptions."""

try:
    from typing import Self
except ImportError:
    # py3.10 and lower
    from typing_extensions import Self


import raimad as rai

class RectWire(rai.Compo):
    """A rectangle defined by two points and a thickness."""

    browser_tags = ["builtin", "polygon"]

    class Options:
        p1 = rai.Option('First point', browser_default=(0, 0))
        p2 = rai.Option('Second point', browser_default=(10, 10))
        width = rai.Option('Thickness', browser_default=2)

    def _make(
            self,
            p1: 'rai.typing.Point',
            p2: 'rai.typing.Point',
            width: float,
            ) -> None:

        angle = rai.angle_between(p1, p2)

        step = rai.polar(
            arg=angle + rai.quartercircle,
            mod=width / 2
            )

        self.geoms.update({
            'root': [
                [
                    (p1[0] + step[0], p1[1] + step[1]),
                    (p1[0] - step[0], p1[1] - step[1]),
                    (p2[0] - step[0], p2[1] - step[1]),
                    (p2[0] + step[0], p2[1] + step[1]),
                    ]
                ]
            })

    @classmethod
    def from_points(
            cls,
            p1: 'rai.typing.Point',
            p2: 'rai.typing.Point',
            width: float,
            ) -> Self:
        return cls(p1, p2, width)

    @classmethod
    def from_polar(
            cls,
            p1: 'rai.typing.Point',
            angle: float,
            length: float,
            width: float,
            ) -> Self:
        step = rai.polar(angle, length)
        p2 = (
            p1[0] + step[0],
            p1[1] + step[1],
            )
        return cls(p1, p2, width)




"""mark.py: home to the Mark class."""
import raimad as rai

class Mark(rai.Annotation):
    """
    Mark annotation.

    Mark objects are meant to exist inside the `Marks` nested
    class of Compos,
    and serve as annotations that better explain what are the
    marks that this compo defines.

    TODO reference boundpoint
    """

    def __init__(self, desc: str):
        self.desc = desc
        self.name = 'this should be set in compo.__init_subclass_'


